""" Tests for tree construction """

import copy
import pytest
import uuid
from le_utils.constants import licenses
from ricecooker.classes.nodes import *
from ricecooker.classes.files import *
from ricecooker.classes.licenses import *
from ricecooker.exceptions import InvalidNodeException


""" *********** CHANNEL FIXTURES *********** """
@pytest.fixture
def domain_namespace():
    return "learningequality.org"

@pytest.fixture
def source_id():
    return "sample-channel"

@pytest.fixture
def channel_domain_namespace(domain_namespace):
    return uuid.uuid5(uuid.NAMESPACE_DNS, domain_namespace)

@pytest.fixture
def channel_node_id(channel_domain_namespace, source_id):
    return uuid.uuid5(channel_domain_namespace, source_id)

@pytest.fixture
def channel_content_id(channel_domain_namespace, channel_node_id):
    return uuid.uuid5(channel_domain_namespace, channel_node_id.hex)

@pytest.fixture
def channel(domain_namespace, source_id):
    channel = ChannelNode(
        source_id=source_id,
        source_domain=domain_namespace,
        title='Channel',
        language='en'
    )
    return channel

@pytest.fixture
def invalid_channel(domain_namespace, source_id):
    channel = ChannelNode(
        source_id=source_id,
        source_domain=domain_namespace,
        title='Channel',
        language='en'
    )
    channel.source_id = None
    return channel


""" *********** TOPIC FIXTURES *********** """
@pytest.fixture
def topic_id():
    return "topic-id"

@pytest.fixture
def topic_content_id(channel_domain_namespace, topic_id):
    return uuid.uuid5(channel_domain_namespace, topic_id)

@pytest.fixture
def topic_node_id(channel_node_id, topic_content_id):
    return uuid.uuid5(channel_node_id, topic_content_id.hex)

@pytest.fixture
def topic(topic_id):
    return TopicNode(topic_id, "Topic")

@pytest.fixture
def invalid_topic(topic_id):
    topic = TopicNode(topic_id, "Topic")
    topic.title = None
    return topic

""" *********** DOCUMENT FIXTURES *********** """
@pytest.fixture
def document_id():
    return "document-id"

@pytest.fixture
def document_content_id(channel_domain_namespace, document_id):
    return uuid.uuid5(channel_domain_namespace, document_id)

@pytest.fixture
def document_node_id(topic_node_id, document_content_id):
    return uuid.uuid5(topic_node_id, document_content_id.hex)

@pytest.fixture
def document_file():
    return DocumentFile("document_path")

@pytest.fixture
def thumbnail_path():
    return "thumbnail path"

@pytest.fixture
def copyright_holder():
    return "Copyright Holder"

@pytest.fixture
def license_name():
    return licenses.PUBLIC_DOMAIN

@pytest.fixture
def document(document_id, document_file, thumbnail_path, copyright_holder, license_name):
    node = DocumentNode(document_id, "Document", licenses.CC_BY, thumbnail=thumbnail_path)
    node.add_file(document_file)
    node.set_license(license_name, copyright_holder=copyright_holder)
    return node

@pytest.fixture
def invalid_document(document_file):
    node = DocumentNode("invalid", "Document", licenses.CC_BY, files=[document_file])
    node.license = None
    return node


""" *********** TREE FIXTURES *********** """
@pytest.fixture
def tree(channel, topic, document):
    topic.add_child(document)
    channel.add_child(topic)
    return channel

@pytest.fixture
def invalid_tree(invalid_channel, invalid_topic, invalid_document):
    invalid_topic.add_child(invalid_document)
    invalid_channel.add_child(invalid_topic)
    return invalid_channel

@pytest.fixture
def invalid_tree_2(channel, topic, invalid_document):
    channel_copy = copy.deepcopy(channel)
    topic_copy = copy.deepcopy(topic)
    topic_copy.add_child(invalid_document)
    channel_copy.add_child(topic_copy)
    return channel_copy

""" *********** CONTENT NODE TESTS *********** """
def test_nodes_initialized(channel, topic, document):
    assert channel
    assert topic
    assert document

def test_add_child(tree, topic, document):
    assert tree.children[0] == topic, "Channel should have topic child node"
    assert tree.children[0].children[0] == document, "Topic should have a document child node"

def test_ids(tree, channel_node_id, channel_content_id, topic_content_id, topic_node_id, document_content_id, document_node_id):
    channel = tree
    topic = tree.children[0]
    document = topic.children[0]

    assert channel.get_content_id() == channel_content_id, "Channel content id should be {}".format(channel_content_id)
    assert channel.get_node_id() == channel_node_id, "Channel node id should be {}".format(channel_node_id)
    assert topic.get_content_id() == topic_content_id, "Topic content id should be {}".format(topic_content_id)
    assert topic.get_node_id() == topic_node_id, "Topic node id should be {}".format(topic_node_id)
    assert document.get_content_id() == document_content_id, "Document content id should be {}".format(document_content_id)
    assert document.get_node_id() == document_node_id, "Document node id should be {}".format(document_node_id)

def test_add_file(document, document_file):
    test_files = list(filter(lambda f: isinstance(f, DocumentFile), document.files))
    assert any(test_files), "Document must have at least one file"
    assert test_files[0] == document_file, "Document file was not added correctly"

def test_thumbnail(topic, document, thumbnail_path):
    assert document.has_thumbnail(), "Document must have a thumbnail"
    assert not topic.has_thumbnail(), "Topic must not have a thumbnail"
    assert any(filter(lambda f: f.path == thumbnail_path, document.files)), "Document is missing a thumbnail with path {}".format(thumbnail_path)

def test_count(tree):
    assert tree.count() == 2, "Channel should have 2 descendants"

def test_get_non_topic_descendants(tree, document):
    assert tree.get_non_topic_descendants() == [document], "Channel should only have 1 non-topic descendant"

def test_licenses(channel, topic, document, license_name, copyright_holder):
    assert isinstance(document.license, License), "Document should have a license object"
    assert document.license.license_id == license_name, "Document license should have public domain license"
    assert document.license.copyright_holder == copyright_holder, "Document license should have copyright holder set to {}".format(copyright_holder)
    assert not channel.license, "Channel should not have a license"
    assert not topic.license, "Topic should not have a license"

def test_tree_validation(tree, invalid_tree, invalid_tree_2):
    assert tree.test_tree(), "Valid tree should pass validation"

    try:
        invalid_tree.test_tree()
        assert False, "Invalid tree should fail validation"
    except InvalidNodeException:
        pass

    try:
        invalid_tree_2.test_tree()
        assert False, "Invalid tree should fail validation"
    except InvalidNodeException:
        pass




# Node models to represent channel's tree
from __future__ import unicode_literals

import hashlib
import json
import os
import shutil
import tempfile
import zipfile
from subprocess import CalledProcessError
import youtube_dl
from le_utils.constants import content_kinds,file_formats, format_presets, exercises, languages
from .. import config
from ..exceptions import UnknownFileTypeError
from cachecontrol.caches.file_cache import FileCache
from le_utils.constants import file_formats, format_presets, exercises
from pressurecooker.encodings import get_base64_encoding, write_base64_to_file
from pressurecooker.images import create_tiled_image
from pressurecooker.videos import extract_thumbnail_from_video, guess_video_preset_by_resolution, compress_video, VideoCompressionError
from requests.exceptions import MissingSchema, HTTPError, ConnectionError, InvalidURL, InvalidSchema

from .. import config
from ..exceptions import UnknownFileTypeError

# Cache for filenames
FILECACHE = FileCache(config.FILECACHE_DIRECTORY, forever=True)
HTTP_CAUGHT_EXCEPTIONS = (HTTPError, ConnectionError, InvalidURL, UnicodeDecodeError, UnicodeError, InvalidSchema, IOError, AssertionError)

def generate_key(action, path_or_id, settings=None, default=" (default)"):
    """ generate_key: generate key used for caching
        Args:
            action (str): how video is being processed (e.g. COMPRESSED or DOWNLOADED)
            path_or_id (str): path to video or youtube_id
            settings (dict): settings for compression or downloading passed in by user
            default (str): if settings are None, default to this extension (avoid overwriting keys)
        Returns: filename
    """
    settings = " {}".format(str(sorted(settings.items()))) if settings else default
    return "{}: {}{}".format(action.upper(), path_or_id, settings)

def download(path, default_ext=None):
    """ download: downloads file
        Args: None
        Returns: filename
    """
    key = "DOWNLOAD:{}".format(path)
    if not config.UPDATE and FILECACHE.get(key):
        return FILECACHE.get(key).decode('utf-8')

    config.LOGGER.info("\tDownloading {}".format(path))

    # Write file to temporary file
    with tempfile.TemporaryFile() as tempf:
        hash = write_and_get_hash(path, tempf)
        tempf.seek(0)

        # Get extension of file or default if none found
        extension = os.path.splitext(path)[1][1:].lower()
        if extension not in [key for key, value in file_formats.choices]:
            if default_ext:
                extension = default_ext
            else:
                raise IOError("No extension found: {}".format(path))

        filename = '{0}.{ext}'.format(hash.hexdigest(), ext=extension)

        copy_file_to_storage(filename, tempf)

        FILECACHE.set(key, bytes(filename, "utf-8"))

        return filename

def download_and_convert_video(path, default_ext=file_formats.MP4, ffmpeg_settings=None):
    """ download: downloads file
        Args: None
        Returns: filename
    """
    ffmpeg_settings = ffmpeg_settings or {}
    key = "DOWNLOAD:{}".format(path)
    if not config.UPDATE and FILECACHE.get(key):
        return FILECACHE.get(key).decode('utf-8')

    config.LOGGER.info("\tDownloading {}".format(path))

    # Get extension of file or default if none found
    extension = os.path.splitext(path)[1][1:].lower()
    converted_path = None

    # Convert video to temporary file
    with tempfile.NamedTemporaryFile(suffix=".{}".format(extension), delete=False) as tempf:
        # Write unsupported video to temporary file
        write_and_get_hash(path, tempf)
        tempf.seek(0)

        # Compress video into mp4 file
        path, _ext = os.path.splitext(tempf.name)
        converted_path = "{}.{}".format(path, file_formats.MP4)
        tempf.close()
        compress_video(tempf.name, converted_path, overwrite=True, **ffmpeg_settings)
        os.unlink(tempf.name)

    # Write file to temporary file
    with tempfile.TemporaryFile() as tempf:
        hash = write_and_get_hash(converted_path, tempf)
        tempf.seek(0)

        filename = '{0}.{ext}'.format(hash.hexdigest(), ext=file_formats.MP4)

        copy_file_to_storage(filename, tempf)
        os.unlink(converted_path)

        FILECACHE.set(key, bytes(filename, "utf-8"))

        return filename


def write_and_get_hash(path, write_to_file, hash=None):
    """ write_and_get_hash: write file
        Args: None
        Returns: Hash of file's contents
    """
    hash = hash or hashlib.md5()
    try:
        # Access path
        r = config.DOWNLOAD_SESSION.get(path, stream=True)
        r.raise_for_status()
        for chunk in r:
            write_to_file.write(chunk)
            hash.update(chunk)

    except (MissingSchema, InvalidSchema):
        # If path is a local file path, try to open the file (generate hash if none provided)
        with open(path, 'rb') as fobj:
            for chunk in iter(lambda: fobj.read(2097152), b""):
                write_to_file.write(chunk)
                hash.update(chunk)

    assert write_to_file.tell() > 0, "File failed to write (corrupted)."

    return hash


def copy_file_to_storage(filename, srcfile, delete_original=False):
    # Some files might have been closed, so only filepath will work
    if isinstance(srcfile, str):
        srcfile = open(srcfile, 'rb')

    # Write file to local storage
    with open(config.get_storage_path(filename), 'wb') as destf:
        if delete_original:
            shutil.move(srcfile.name, destf.name)
        else:
            shutil.copyfileobj(srcfile, destf)


def get_hash(filepath):
    file_hash = hashlib.md5()
    with open(filepath, 'rb') as fobj:
        for chunk in iter(lambda: fobj.read(2097152), b""):
            file_hash.update(chunk)
    return file_hash.hexdigest()


def compress_video_file(filename, ffmpeg_settings):
    ffmpeg_settings = ffmpeg_settings or {}
    key = generate_key("COMPRESSED", filename, settings=ffmpeg_settings, default=" (default compression)")

    if not config.UPDATE and FILECACHE.get(key):
        return FILECACHE.get(key).decode('utf-8')

    config.LOGGER.info("\t--- Compressing {}".format(filename))

    tempf = tempfile.NamedTemporaryFile(suffix=".{}".format(file_formats.MP4), delete=False)
    tempf.close() # Need to close so pressure cooker can write to file
    compress_video(config.get_storage_path(filename), tempf.name, overwrite=True, **ffmpeg_settings)
    filename = "{}.{}".format(get_hash(tempf.name), file_formats.MP4)

    copy_file_to_storage(filename, tempf.name)
    os.unlink(tempf.name)
    FILECACHE.set(key, bytes(filename, "utf-8"))
    return filename

def download_from_web(web_url, download_settings, file_format=file_formats.MP4, ext="", download_ext=""):
    key = generate_key("DOWNLOADED", web_url, settings=download_settings)
    if not config.UPDATE and FILECACHE.get(key):
        return FILECACHE.get(key).decode('utf-8')

    # Get hash of web_url to act as temporary storage name
    url_hash = hashlib.md5()
    url_hash.update(web_url.encode('utf-8'))
    destination_path = os.path.join(tempfile.gettempdir(), "{}{ext}".format(url_hash.hexdigest(), ext=ext))
    download_settings["outtmpl"] = destination_path
    try:
        os.remove(destination_path)
    except Exception:
        pass

    with youtube_dl.YoutubeDL(download_settings) as ydl:
        ydl.download([web_url])
        destination_path += download_ext
        filename = "{}.{}".format(get_hash(destination_path), file_format)

        # Write file to local storage
        with open(destination_path, "rb") as dlf, open(config.get_storage_path(filename), 'wb') as destf:
            shutil.copyfileobj(dlf, destf)

        FILECACHE.set(key, bytes(filename, "utf-8"))
        return filename


class ThumbnailPresetMixin(object):

    def get_preset(self):
        thumbnail_preset = self.node.get_thumbnail_preset()
        if thumbnail_preset is None:
            UnknownFileTypeError("Thumbnails are not supported for node kind.")
        return thumbnail_preset


class File(object):
    original_filename = None
    node = None
    error = None
    default_ext = None
    filename = None
    language = None
    assessment_item = None
    is_primary = False

    def __init__(self, preset=None, language=None, default_ext=None, source_url=None):
        self.preset = preset
        self.set_language(language)
        self.default_ext = default_ext or self.default_ext
        self.source_url = source_url

    def set_language(self, language):
        """ Set self.language to internal lang. repr. code from str or Language object. """
        if isinstance(language, str):
            language_obj = languages.getlang(language)
            if language_obj:
                self.language = language_obj.code
            else:
                raise TypeError("Language code {} not found".format(language))
        if isinstance(language, languages.Language):
            self.language = language.code

    def validate(self):
        pass

    def get_preset(self):
        if self.preset:
            return self.preset
        raise NotImplementedError("preset must be set if preset isn't specified when creating File object")

    def get_filename(self):
        return self.filename or self.process_file()

    def truncate_fields(self):
        if self.original_filename and len(self.original_filename) > config.MAX_ORIGINAL_FILENAME_LENGTH:
            config.print_truncate("original_filename", self.node.source_id, self.original_filename)
            self.original_filename = self.original_filename[:config.MAX_ORIGINAL_FILENAME_LENGTH]

        if self.source_url and len(self.source_url) > config.MAX_SOURCE_URL_LENGTH:
            config.print_truncate("file_source_url", self.node.source_id, self.source_url)
            self.source_url = self.source_url[:config.MAX_SOURCE_URL_LENGTH]

    def to_dict(self):
        filename = self.get_filename()

        # If file was successfully downloaded, return dict
        # Otherwise return None
        if filename:
            if os.path.isfile(config.get_storage_path(filename)):
                return {
                    'size' : os.path.getsize(config.get_storage_path(filename)),
                    'preset' : self.get_preset(),
                    'filename' : filename,
                    'original_filename' : self.original_filename,
                    'language' : self.language,
                    'source_url': self.source_url,
                }
            else:
                config.LOGGER.warning("File not found: {}".format(config.get_storage_path(filename)))

        return None

    def process_file(self):
        # Overwrite in subclasses
        pass


class NodeFile(File):
    default_ext = file_formats.JSON
    allowed_formats = [file_formats.JSON]

    def __init__(self, node_metadata, **kwargs):
        self.metadata_json = json.dumps(node_metadata, ensure_ascii=False)
        super(NodeFile, self).__init__(**kwargs)

    def process_file(self):
        assert self.metadata_json, "{} must have node metadata".format(self.__class__.__name__)

        byte_data = self.metadata_json.encode('utf-8')
        with tempfile.TemporaryFile() as temp_file:
            file_hash = hashlib.md5(byte_data)
            temp_file.write(byte_data)

            assert temp_file.tell() > 0, "File failed to write (corrupted)."
            temp_file.seek(0)

            extension = self.default_ext
            file_name = '{0}.{ext}'.format(file_hash.hexdigest(), ext=extension)

            copy_file_to_storage(file_name, temp_file)

            return file_name


class DownloadFile(File):
    allowed_formats = []

    def __init__(self, path, **kwargs):
        self.path = path.strip()
        super(DownloadFile, self).__init__(**kwargs)

    def validate(self):
        assert self.path, "{} must have a path".format(self.__class__.__name__)
        _basename, ext = os.path.splitext(self.path)
        plain_ext = ext.lstrip('.')
        # don't validate for single-digit extension, or no extension
        if len(plain_ext) > 1:
            assert plain_ext in self.allowed_formats, "{} must have one of the following extensions: {} (instead, got '{}' from '{}')".format(self.__class__.__name__, self.allowed_formats, plain_ext, self.path)

    def process_file(self):
        try:
            self.filename = download(self.path, default_ext=self.default_ext)
            config.LOGGER.info("\t--- Downloaded {}".format(self.filename))
            return self.filename
        # Catch errors related to reading file path and handle silently
        except HTTP_CAUGHT_EXCEPTIONS as err:
            self.error = err
            config.FAILED_FILES.append(self)

    def __str__(self):
        return self.path

class ThumbnailFile(ThumbnailPresetMixin, DownloadFile):
    default_ext = file_formats.PNG
    allowed_formats = [file_formats.JPG, file_formats.JPEG, file_formats.PNG]

class AudioFile(DownloadFile):
    default_ext = file_formats.MP3
    allowed_formats = [file_formats.MP3]
    is_primary = True

    def get_preset(self):
        return self.preset or format_presets.AUDIO

class DocumentFile(DownloadFile):
    default_ext = file_formats.PDF
    allowed_formats = [file_formats.PDF]
    is_primary = True

    def get_preset(self):
        return self.preset or format_presets.DOCUMENT

class HTMLZipFile(DownloadFile):
    default_ext = file_formats.HTML5
    allowed_formats = [file_formats.HTML5]
    is_primary = True

    def get_preset(self):
        return self.preset or format_presets.HTML5_ZIP

    def process_file(self):
        self.filename = super(HTMLZipFile, self).process_file()

        # make sure index.html exists
        with zipfile.ZipFile(config.get_storage_path(self.filename)) as zf:
            try:
                info = zf.getinfo('index.html')
            except KeyError:
                raise IOError("HTML zip must have an `index.html` file at topmost level")

        return self.filename

    def validate(self):
        super(HTMLZipFile, self).validate()

class ExtractedVideoThumbnailFile(ThumbnailFile):

    def process_file(self):
        self.filename = self.derive_thumbnail()
        config.LOGGER.info("\t--- Extracted thumbnail {}".format(self.filename))
        return self.filename

    def derive_thumbnail(self):
        key = "EXTRACTED: {}".format(self.path)
        if not config.UPDATE and FILECACHE.get(key):
            return FILECACHE.get(key).decode('utf-8')

        config.LOGGER.info("\t--- Extracting thumbnail from {}".format(self.path))
        tempf = tempfile.NamedTemporaryFile(suffix=".{}".format(file_formats.PNG), delete=False)
        tempf.close()
        extract_thumbnail_from_video(self.path, tempf.name, overwrite=True)
        filename = "{}.{}".format(get_hash(tempf.name), file_formats.PNG)

        copy_file_to_storage(filename, tempf.name)
        os.unlink(tempf.name)
        FILECACHE.set(key, bytes(filename, "utf-8"))
        return filename

class VideoFile(DownloadFile):
    default_ext = file_formats.MP4
    allowed_formats = [file_formats.MP4, 'avi', 'mov', 'mpg', 'wmv', 'swf', 'webm', 'mkv', 'flv']
    is_primary = True

    def __init__(self, path, ffmpeg_settings=None, **kwargs):
        self.ffmpeg_settings = ffmpeg_settings
        super(VideoFile, self).__init__(path, **kwargs)

    def get_preset(self):
        return self.preset or guess_video_preset_by_resolution(config.get_storage_path(self.filename))

    def process_unsupported_video_file(self):
        try:
            self.filename = download_and_convert_video(self.path, ffmpeg_settings=self.ffmpeg_settings)
            config.LOGGER.info("\t--- Downloaded and converted {}".format(self.filename))
            return self.filename
        except HTTP_CAUGHT_EXCEPTIONS as err:
            self.error = err
            config.FAILED_FILES.append(self)

    def process_file(self):
        try:
            # Handle videos that don't have a .mp4 extension
            _, ext = os.path.splitext(self.path)
            if ext and not ext.endswith(self.default_ext):
                self.filename = self.process_unsupported_video_file()
                return self.filename
            else:
                # Get copy of video before compression (if specified)
                self.filename = super(VideoFile, self).process_file()

                # Compress the video if compress flag is set or ffmpeg settings were given
                if self.filename and (self.ffmpeg_settings or config.COMPRESS):
                    self.filename = compress_video_file(self.filename, self.ffmpeg_settings)
                    config.LOGGER.info("\t--- Compressed {}".format(self.filename))
                return self.filename

        # Catch errors related to ffmpeg and handle silently
        except (BrokenPipeError, CalledProcessError, IOError, VideoCompressionError) as err:
            self.error = err
            config.FAILED_FILES.append(self)

class WebVideoFile(File):
    is_primary = True
    # In future, look into postprocessors and progress_hooks
    def __init__(self, web_url, download_settings=None, high_resolution=True, maxheight=None, **kwargs):
        self.web_url = web_url
        self.download_settings = download_settings or {}
        if "format" not in self.download_settings:
            maxheight = maxheight or (720 if high_resolution else 480)
            self.download_settings['format'] = "bestvideo[height<={maxheight}][ext=mp4]+bestaudio[ext=m4a]/best[height<={maxheight}][ext=mp4]".format(maxheight=maxheight)
            # self.download_settings['recodevideo'] = file_formats.MP4

        super(WebVideoFile, self).__init__(**kwargs)

    def get_preset(self):
        return self.preset or guess_video_preset_by_resolution(config.get_storage_path(self.filename))

    def process_file(self):
        try:
            self.filename = download_from_web(self.web_url, self.download_settings, ext=".{}".format(file_formats.MP4))
            config.LOGGER.info("\t--- Downloaded (YouTube) {}".format(self.filename))
            return self.filename
        except youtube_dl.utils.DownloadError as err:
            self.error = str(err)
            config.FAILED_FILES.append(self)

class YouTubeVideoFile(WebVideoFile):
    def __init__(self, youtube_id, **kwargs):
        super(YouTubeVideoFile, self).__init__('http://www.youtube.com/watch?v={}'.format(youtube_id), **kwargs)

class YouTubeSubtitleFile(File):
    """
    Helper class for downloading youtube subtitles.
    Args:
       youtube_id (string): YouTube ID of video (required)
       language (string): internal language id format `{primary_code}` or `{primary_code}-{subcode}` (required) \
                          alternatively, you can provide the language code recognized by YouTube
    """
    def __init__(self, youtube_id, language=None, **kwargs):
        self.youtube_url = 'http://www.youtube.com/watch?v={}'.format(youtube_id)
        if isinstance(language, languages.Language):  # for backward compatibility
            language = language.code
        self.youtube_language = language  # youtube language code (can differ from internal repr.)
        language_obj = languages.getlang(language)   # lookup `language` using internal representation
        # if language_obj not None, we know `language` is a valid language_id in the internal repr.
        if language_obj is None:  # if `language` not found using internal repr.
            language_obj = languages.getlang_by_alpha2(language)  # try to match by two-letter ISO code
            language = language_obj.code   # update `language` argument from internal repr. language_id
        super(YouTubeSubtitleFile, self).__init__(language=language, **kwargs)
        assert self.language, "Subtitles must have a language"

    def get_preset(self):
        return self.preset or format_presets.VIDEO_SUBTITLE

    def process_file(self):
        try:
            self.filename = self.download_subtitle()
            config.LOGGER.info("\t--- Downloaded subtitle {}".format(self.filename))
            return self.filename
        except (FileNotFoundError, youtube_dl.utils.DownloadError):
            self.error = str("Subtitle with langauge {} is not available for {}".format(self.language, self.youtube_url))
            config.FAILED_FILES.append(self)

    def download_subtitle(self):
        settings = {
            'skip_download': True,
            'writesubtitles': True,
            'subtitleslangs': [self.youtube_language],
            'subtitlesformat': "best[ext={}]".format(file_formats.VTT),
            'quiet': True,
            'no_warnings': True
        }
        download_ext = ".{lang}.{ext}".format(lang=self.youtube_language, ext=file_formats.VTT)
        return download_from_web(self.youtube_url, settings, file_format=file_formats.VTT, download_ext=download_ext)

class SubtitleFile(DownloadFile):
    default_ext = file_formats.VTT
    allowed_formats = [file_formats.VTT]

    def __init__(self, path, **kwargs):
        super(SubtitleFile, self).__init__(path, **kwargs)
        assert self.language, "Subtitles must have a language"

    def get_preset(self):
        return self.preset or format_presets.VIDEO_SUBTITLE


class Base64ImageFile(ThumbnailPresetMixin, File):

    def __init__(self, encoding, **kwargs):
        self.encoding = encoding
        super(Base64ImageFile, self).__init__(**kwargs)

    def process_file(self):
        """ process_file: Writes base64 encoding to file
            Args: None
            Returns: filename
        """
        self.filename = self.convert_base64_to_file()
        config.LOGGER.info("\t--- Converted base64 image to {}".format(self.filename))
        return self.filename

    def convert_base64_to_file(self):
        # Get hash of content for cache key
        hashed_content = hashlib.md5()
        hashed_content.update(self.encoding.encode('utf-8'))
        key = "ENCODED: {} (base64 encoded)".format(hashed_content.hexdigest())

        if not config.UPDATE and FILECACHE.get(key):
            return FILECACHE.get(key).decode('utf-8')

        config.LOGGER.info("\tConverting base64 to file")

        extension = get_base64_encoding(self.encoding).group(1)
        assert extension in [file_formats.PNG, file_formats.JPG, file_formats.JPEG], "Base64 files must be images in jpg or png format"

        tempf = tempfile.NamedTemporaryFile(suffix=".{}".format(extension), delete=False)
        tempf.close()
        write_base64_to_file(self.encoding, tempf.name)
        filename = "{}.{}".format(get_hash(tempf.name), file_formats.PNG)

        copy_file_to_storage(filename, tempf.name)
        os.unlink(tempf.name)
        FILECACHE.set(key, bytes(filename, "utf-8"))
        return filename

class _ExerciseBase64ImageFile(Base64ImageFile):
    default_ext = file_formats.PNG

    def get_preset(self):
        return self.preset or format_presets.EXERCISE_IMAGE

    def get_replacement_str(self):
        return self.get_filename() or self.encoding

class _ExerciseImageFile(DownloadFile):
    default_ext = file_formats.PNG

    def get_replacement_str(self):
        return self.get_filename() or self.path

    def get_preset(self):
        return self.preset or format_presets.EXERCISE_IMAGE

class _ExerciseGraphieFile(DownloadFile):
    default_ext = file_formats.GRAPHIE

    def __init__(self, path, **kwargs):
        self.original_filename = path.split("/")[-1].split(os.path.sep)[-1].split(".")[0]
        super(_ExerciseGraphieFile, self).__init__(path, **kwargs)

    def get_preset(self):
        return self.preset or format_presets.EXERCISE_GRAPHIE

    def get_replacement_str(self):
        return self.path.split("/")[-1].split(".")[0] or self.path

    def process_file(self):
        """ download: download a web+graphie file
            Args: None
            Returns: None
        """
        try:
            self.filename = self.generate_graphie_file()
            config.LOGGER.info("\t--- Generated graphie {}".format(self.filename))
            return self.filename
        # Catch errors related to reading file path and handle silently
        except (HTTPError, ConnectionError, InvalidURL, UnicodeDecodeError, UnicodeError, InvalidSchema, IOError) as err:
            self.error = err
            config.FAILED_FILES.append(self)

    def generate_graphie_file(self):
        key = "GRAPHIE: {}".format(self.path)

        if not config.UPDATE and FILECACHE.get(key):
            return FILECACHE.get(key).decode('utf-8')

        # Create graphie file combining svg and json files
        with tempfile.TemporaryFile() as tempf:
            # Initialize hash and files
            delimiter = bytes(exercises.GRAPHIE_DELIMITER, 'UTF-8')
            config.LOGGER.info("\tDownloading graphie {}".format(self.original_filename))


            # Write to graphie file
            hash = write_and_get_hash(self.path + ".svg", tempf)
            tempf.write(delimiter)
            hash.update(delimiter)
            hash = write_and_get_hash(self.path + "-data.json", tempf, hash)
            tempf.seek(0)
            filename = "{}.{}".format(hash.hexdigest(), file_formats.GRAPHIE)

            copy_file_to_storage(filename, tempf)

            FILECACHE.set(key, bytes(filename, "utf-8"))
            return filename


class TiledThumbnailFile(ThumbnailPresetMixin, File):
    allowed_formats = [file_formats.JPG, file_formats.JPEG, file_formats.PNG]

    def __init__(self, source_nodes, **kwargs):
        self.sources = []
        for n in source_nodes:
            images = [f for f in n.files if isinstance(f, ThumbnailFile) and f.get_filename()]
            if len(images) > 0:
                self.sources.append(images[0])
        super(TiledThumbnailFile, self).__init__(**kwargs)

    def process_file(self):
        self.filename = self.generate_tiled_image()
        config.LOGGER.info("\t--- Tiled image {}".format(self.filename))
        return self.filename

    def generate_tiled_image(self):
        num_pictures = 0
        if len(self.sources) >= 4:
            num_pictures = 4
        elif len(self.sources) >= 1:
            num_pictures = 1
        else:
            return None

        images = [config.get_storage_path(f.get_filename()) for f in self.sources[:num_pictures]]
        key = "TILED {}".format("+".join(sorted(images)))
        if not config.UPDATE and FILECACHE.get(key):
            return FILECACHE.get(key).decode('utf-8')

        config.LOGGER.info("\tTiling thumbnail for {}".format(self.node.title))
        with tempfile.NamedTemporaryFile(suffix=".{}".format(file_formats.PNG)) as tempf:
            tempf.close()
            create_tiled_image(images, tempf.name)
            filename = "{}.{}".format(get_hash(tempf.name), file_formats.PNG)

            copy_file_to_storage(filename, tempf.name)
            FILECACHE.set(key, bytes(filename, "utf-8"))
            return filename


# VectorizedVideoFile
# UniversalSubsSubtitleFile

# class UniversalSubsSubtitleFile(SubtitleFile):
#     def __init__(self, us_id, language):
#         response = sess.get("http://usubs.org/api/{}".format(us_id))
#         path = json.loads(response.content)["subtitle_url"]
#         return super(UniversalSubsSubtitleFile, self).__init__(path=path, language=language)



import argparse
import os

from ricecooker.config import LOGGER
from le_utils.constants import content_kinds
from .metadata_provider import path_to_tuple
from .jsontrees import (TOPIC_NODE, VIDEO_NODE, AUDIO_NODE, EXERCISE_NODE,
                        DOCUMENT_NODE, HTML5_NODE)
from .jsontrees import (VIDEO_FILE, AUDIO_FILE, DOCUMENT_FILE, HTML5_FILE,
                        THUMBNAIL_FILE, SUBTITLES_FILE)
from .jsontrees import write_tree_to_json_tree


# LINECOOK CONFIGS
################################################################################
DIR_EXCLUDE_PATTERNS = []
FILE_EXCLUDE_EXTENTIONS = ['.DS_Store', 'Thumbs.db', 'ehthumbs.db', 'ehthumbs_vista.db', '.gitkeep']
FILE_SKIP_PATTENRS = []
FILE_SKIP_THUMBNAILS = []  # global list of paths that correspond to thumbails for other content nodes



# LINECOOK HELPER FUNCTIONS
################################################################################

def chan_path_from_rel_path(rel_path, channeldir):
    """
    Convert `rel_path` form os.walk tuple format to a tuple of directories and
    subdirectories, starting with the `channeldir` folder, e.g.,::

        >>> chan_path_from_rel_path('content/open_stax_zip/Open Stax/Math/Elementary',
                                'content/open_stax_zip/Open Stax')
        'Open Stax/Math/Elementary'
    """
    rel_path_parts = rel_path.split(os.path.sep)
    dirs_before_channeldir = channeldir.split(os.path.sep)[:-1]
    channel_chan_path = []  # path relative to channel root, inclusive
    for idx, part in enumerate(rel_path_parts):
        if idx < len(dirs_before_channeldir) and dirs_before_channeldir[idx]==part:
            continue
        else:
            channel_chan_path.append(part)
    chan_path = os.path.join(*channel_chan_path)
    return chan_path

def rel_path_from_chan_path(chan_path, channeldir, windows=False):
    """
    Convert `chan_path` as obtained from a metadata provider into a `rel_path`
    suitable for accessing the file from the current working directory, e.g.,
    >>> rel_path_from_chan_path('Open Stax/Math', 'content/open_stax_zip/Open Stax')
    'content/open_stax_zip/Open Stax/Math'
    """
    if windows:
        chan_path_list = chan_path.split('\\')
    else:
        chan_path_list = chan_path.split('/')
    chan_path_list.pop(0)  # remove the channel root dir
    rel_path = os.path.join(channeldir, *chan_path_list)
    return rel_path

def get_topic_for_path(channel, chan_path_tuple):
    """
    Given channel (dict) that contains a hierary of TopicNode dicts, we use the
    walk the path given in `chan_path_tuple` to find the corresponding TopicNode.
    """
    assert chan_path_tuple[0] == channel['dirname'], 'Wrong channeldir'
    chan_path_list = list(chan_path_tuple)
    chan_path_list.pop(0)    # skip the channel name

    if len(chan_path_list) == 0:
        return channel

    current = channel
    for subtopic in chan_path_list:
        current = list(filter(lambda d: 'dirname' in d and d['dirname'] == subtopic, current['children']))[0]
    return current



# LINECOOK BUILD JSON TREE
################################################################################

def filter_filenames(filenames):
    """
    Skip files with extentions in `FILE_EXCLUDE_EXTENTIONS` and filenames that
    contain `FILE_SKIP_PATTENRS`.
    """
    filenames_cleaned = []
    for filename in filenames:
        keep = True
        for pattern in FILE_EXCLUDE_EXTENTIONS:
            if filename.endswith(pattern):
                keep = False
        for pattern in FILE_SKIP_PATTENRS:   # This will reject exercises...
            if pattern in filename:
                keep = False
        if keep:
            filenames_cleaned.append(filename)
    return filenames_cleaned

def filter_thumbnail_files(chan_path, filenames, metadata_provider):
    """
    We don't want to create `ContentNode` from thumbnail files.
    """
    thumbnail_files_to_skip = metadata_provider.get_thumbnail_paths()
    filenames_cleaned = []
    for filename in filenames:
        keep = True
        chan_filepath = os.path.join(chan_path, filename)
        chan_filepath_tuple = path_to_tuple(chan_filepath)
        if chan_filepath_tuple in thumbnail_files_to_skip:
            keep = False
        if keep:
            filenames_cleaned.append(filename)
    return filenames_cleaned

def keep_folder(raw_path):
    """
    Keep only folders that don't contain patterns in `DIR_EXCLUDE_PATTERNS`.
    """
    keep = True
    for pattern in DIR_EXCLUDE_PATTERNS:
        if pattern in raw_path:
            LOGGER.debug('rejecting', raw_path)
            keep = False
    return keep

def process_folder(channel, rel_path, filenames, metadata_provider):
    """
    Create `ContentNode`s from each file in this folder and the node to `channel`
    under the path `rel_path`.
    """
    LOGGER.debug('IN process_folder ' + str(rel_path) + '     ' + str(filenames))
    if not keep_folder(rel_path):
        return

    chan_path = chan_path_from_rel_path(rel_path, metadata_provider.channeldir)
    chan_path_tuple = path_to_tuple(chan_path)
    chan_path_list = list(chan_path_tuple)
    LOGGER.debug('chan_path_list=' + str(chan_path_list))

    # FIND THE CONTAINING NODE (channel or topic)
    if len(chan_path_list) == 1:
        # CASE CHANNEL ROOT: `rel_path` points to `channeldir`
        # No need to create a topic node here since channel already exists
        containing_node = channel  # attach content nodes in filenames directly to channel

    else:
        # CASE TOPIC FOLDER: `rel_path` points to a channelroot subfolder (a.k.a TopicNode)
        dirname = chan_path_list.pop()  # name of the folder (used as ID for internal lookup)
        topic_parent_node = get_topic_for_path(channel, chan_path_list)

        # read topic metadata to get title and description for the TopicNode
        topic_metadata = metadata_provider.get(chan_path_tuple)
        thumbnail_chan_path =  topic_metadata.get('thumbnail_chan_path', None)
        if thumbnail_chan_path:
            thumbnail_rel_path = rel_path_from_chan_path(thumbnail_chan_path, metadata_provider.channeldir)
        else:
            thumbnail_rel_path = None
        # create TopicNode for this folder
        topic = dict(
            kind=TOPIC_NODE,
            dirname=dirname,
            source_id='sourceid:' + rel_path,
            title=topic_metadata.get('title', dirname),
            description=topic_metadata.get('description', None),
            author=topic_metadata.get('author', None),
            language=topic_metadata.get('language', None),
            license=topic_metadata.get('license', None),
            thumbnail=thumbnail_rel_path,
            children=[],
        )
        topic_parent_node['children'].append(topic)
        containing_node = topic  # attach content nodes in filenames to the newly created topic

    # filter filenames
    filenames_cleaned = filter_filenames(filenames)
    filenames_cleaned2 = filter_thumbnail_files(chan_path, filenames_cleaned, metadata_provider)

    # PROCESS FILES
    for filename in filenames_cleaned2:
        chan_filepath = os.path.join(chan_path, filename)
        chan_filepath_tuple = path_to_tuple(chan_filepath)
        metadata = metadata_provider.get(chan_filepath_tuple)
        node = make_content_node(metadata_provider.channeldir, rel_path, filename, metadata)
        containing_node['children'].append(node)  # attach content node to containing_node


def build_ricecooker_json_tree(args, options, metadata_provider, json_tree_path):
    """
    Download all categories, subpages, modules, and resources from open.edu.
    """
    LOGGER.info('Starting to build the ricecooker_json_tree')

    channeldir = args['channeldir']
    if channeldir.endswith(os.path.sep):
        channeldir.rstrip(os.path.sep)
    channelparentdir, channeldirname = os.path.split(channeldir)
    channelparentdir, channeldirname = os.path.split(channeldir)

    # Ricecooker tree
    channel_info = metadata_provider.get_channel_info()
    thumbnail_chan_path =  channel_info.get('thumbnail_chan_path', None)
    if thumbnail_chan_path:
        thumbnail_rel_path = rel_path_from_chan_path(thumbnail_chan_path, metadata_provider.channeldir)
    else:
        thumbnail_rel_path = None

    ricecooker_json_tree = dict(
        dirname=channeldirname,
        title=channel_info['title'],
        description=channel_info['description'],
        source_domain=channel_info['source_domain'],
        source_id=channel_info['source_id'],
        language=channel_info['language'],
        thumbnail=thumbnail_rel_path,
        children=[],
    )
    channeldir = args['channeldir']
    content_folders = sorted(os.walk(channeldir))

    # MAIN PROCESSING OF os.walk OUTPUT
    ############################################################################
    # TODO(ivan): figure out all the implications of the
    # _ = content_folders.pop(0)  # Skip over channel folder because handled above
    for rel_path, _subfolders, filenames in content_folders:
        LOGGER.info('processing folder ' + str(rel_path))

        # IMPLEMENTATION DETAIL:
        #   - `filenames` contains real files in the `channeldir` folder
        #   - `exercises_filenames` contains virtual files whose sole purpse is to set the
        #     order of nodes within a given topic. Since alphabetical order is used to
        #     walk the files in the `channeldir`, we must "splice in" the exercises here
        if metadata_provider.has_exercises():
            dir_chan_path = chan_path_from_rel_path(rel_path, metadata_provider.channeldir)
            dir_path_tuple = path_to_tuple(dir_chan_path)
            exercises_filenames = metadata_provider.get_exercises_for_dir(dir_path_tuple)
            filenames.extend(exercises_filenames)

        sorted_filenames = sorted(filenames)
        process_folder(ricecooker_json_tree, rel_path, sorted_filenames, metadata_provider)

    # Write out ricecooker_json_tree.json
    write_tree_to_json_tree(json_tree_path, ricecooker_json_tree)
    LOGGER.info('Folder hierarchy walk result stored in ' + json_tree_path)


def make_content_node(channeldir, rel_path, filename, metadata):
    """
    Create ContentNode based on the file extention and metadata provided.
    """
    file_key, file_ext = os.path.splitext(filename)
    ext = file_ext[1:]
    kind = None
    if ext in content_kinds.MAPPING:
        kind = content_kinds.MAPPING[ext]   # guess what kind based on file extension
    elif 'questions' in metadata:
        kind = content_kinds.EXERCISE
    else:
        raise ValueError('Could not find kind for extension ' + str(ext) + ' in content_kinds.MAPPING')

    # Extract metadata fields
    source_id = metadata.get('source_id', None)
    if source_id is None:
        source_id = metadata['chan_path']

    filepath = os.path.join(rel_path, filename)
    title = metadata['title']
    description = metadata.get('description', None)
    author = metadata.get('author', None)
    lang =  metadata.get('language', None)
    license_dict = metadata.get('license', None)
    thumbnail_chan_path =  metadata.get('thumbnail_chan_path', None)
    if thumbnail_chan_path:
        thumbnail_rel_path = rel_path_from_chan_path(thumbnail_chan_path, channeldir)
    else:
        thumbnail_rel_path = None

    if kind == VIDEO_NODE:
        content_node = dict(
            kind=VIDEO_NODE,
            source_id=source_id,
            title=title,
            author=author,
            description=description,
            language=lang,
            license=license_dict,
            derive_thumbnail=True,  # video-specific option
            thumbnail=thumbnail_rel_path,
            files=[{'file_type':VIDEO_FILE, 'path':filepath, 'language':lang}], # ffmpeg_settings={"crf": 24},
        )

    elif kind == AUDIO_NODE:
        content_node = dict(
            kind=AUDIO_NODE,
            source_id=source_id,
            title=title,
            author=author,
            description=description,
            language=lang,
            license=license_dict,
            thumbnail=thumbnail_rel_path,
            files=[{'file_type':AUDIO_FILE, 'path':filepath, 'language':lang}],
        )

    elif kind == DOCUMENT_NODE:
        content_node = dict(
            kind=DOCUMENT_NODE,
            source_id=source_id,
            title=title,
            author=author,
            description=description,
            language=lang,
            license=license_dict,
            thumbnail=thumbnail_rel_path,
            files=[{'file_type':DOCUMENT_FILE, 'path':filepath, 'language':lang}],
        )

    elif kind == HTML5_NODE:
        content_node = dict(
            kind=HTML5_NODE,
            source_id=source_id,
            title=title,
            author=author,
            description=description,
            language=lang,
            license=license_dict,
            thumbnail=thumbnail_rel_path,
            files=[{'file_type':HTML5_FILE, 'path':filepath, 'language':lang}],
        )

    elif kind == EXERCISE_NODE:
        content_node = dict(
            kind=EXERCISE_NODE,
            source_id=source_id,
            title=title,
            author=author,
            description=description,
            language=lang,
            license=license_dict,
            exercise_data=metadata['exercise_data'],
            questions=metadata['questions'],
            thumbnail=thumbnail_rel_path,
            files=[],
        )

    else:
        raise ValueError('Not implemented case for kind ' + str(kind))

    return content_node



# AUTOMATIC REMOVAL OF TRAILING SLASHES FOR chenneldir
################################################################################

class NonFolderError(Exception):
    pass

class FolderExistsAction(argparse.Action):
    """
    Custom argparse action: verify the argument to be a folder (directory).
    The action will strip off trailing slashes from the folder's name.
    """
    def verify_folder_existence(self, folder_name):
        if not os.path.isdir(folder_name):
            message = 'ERROR: {0} is not a folder'.format(folder_name)
            raise NonFolderError(message)
        folder_name = folder_name.rstrip(os.sep)
        return folder_name

    def __call__(self, parser, namespace, values, option_string=None):
        if type(values) == list:
            folders = map(self.verify_folder_existence, values)
        else:
            folders = self.verify_folder_existence(values)
        setattr(namespace, self.dest, folders)




import json
import os
try: # to support Python 2.x.
    input = raw_input
except NameError:
    pass

from requests.exceptions import HTTPError

from ricecooker import config


def get_env(envvar):
    """
    Reads an environment variable `envvar` if it is defined or returns None.
    """
    if envvar not in os.environ:
        return None
    else:
        return os.environ[envvar]

def get_content_curation_token(args_token):
    """
    Get the token through one of four possible ways. Input `args_token` can be
    1. path to a token-containing file (path)
    2. actual token (str) in which case there's nothing to get just pass along
    3. `#` (default value when no --token is given on command line)
    3a. if environment variable CONTENT_CURATION_TOKEN exists, we'll use that
    3b. else we prompt the user interactively
    """
    if args_token != "#":                               # retrieval methods 1, 2
        if os.path.isfile(args_token):
            with open(args_token, 'r') as fobj:
                return fobj.read().strip()
        else:
            return args_token
    else:                                               # retrieval strategies 3
        token = get_env('STUDIO_TOKEN') or get_env('CONTENT_CURATION_TOKEN')
        if token is not None:
            return token                                # 3a
        else:
            return prompt_token(config.DOMAIN)          # 3b

def prompt_token(domain):
    """
    Prompt user to enter content curation server authentication token.
    Args: domain (str): domain to authenticate user
    Returns: token
    """
    token = input("\nEnter content curation server token ('q' to quit): ").lower()
    if token == 'q':
        sys.exit()
    else:
        return token.strip()

# SUSHI_BAR_TOKEN = get_env('SUSHI_BAR_TOKEN')  # TODO in near future


# License models

from ..exceptions import UnknownLicenseError
from .. import config
from le_utils.constants import licenses


def get_license(license_id, copyright_holder=None, description=None):
    if license_id == licenses.CC_BY:
        return CC_BYLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.CC_BY_SA:
        return CC_BY_SALicense(copyright_holder=copyright_holder)
    elif license_id == licenses.CC_BY_ND:
        return CC_BY_NDLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.CC_BY_NC:
        return CC_BY_NCLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.CC_BY_NC_SA:
        return CC_BY_NC_SALicense(copyright_holder=copyright_holder)
    elif license_id == licenses.CC_BY_NC_ND:
        return CC_BY_NC_NDLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.ALL_RIGHTS_RESERVED:
        return AllRightsLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.PUBLIC_DOMAIN:
        return PublicDomainLicense(copyright_holder=copyright_holder)
    elif license_id == licenses.SPECIAL_PERMISSIONS:
        return SpecialPermissionsLicense(copyright_holder=copyright_holder, description=description)
    else:
        raise UnknownLicenseError("{} is not a valid license id. (Valid license are {})".format(license_id, [l[0] for l in licenses.choices]))


class License(object):
    license_id = None # (str): content's license based on le_utils.constants.licenses
    copyright_holder = None # (str): name of person or organization who owns license (optional)
    description = None # (str): description of the license (optional)
    require_copyright_holder = True

    def __init__(self, copyright_holder=None, description=None):
        self.copyright_holder = copyright_holder or ""
        self.description = description

    def get_id(self):
        return self.license_id

    def validate(self):
        assert not self.require_copyright_holder or self.copyright_holder != "", "Assertion Failed: {} License requires a copyright holder".format(self.license_id)
        assert isinstance(self.copyright_holder, str), "Assertion Failed: Copyright holder must be a string"

    def truncate_fields(self):
        if self.description and len(self.description) > config.MAX_LICENSE_DESCRIPTION_LENGTH:
            config.print_truncate("license_description", self.license_id, self.description)
            self.description = self.description[:config.MAX_LICENSE_DESCRIPTION_LENGTH]

        if self.copyright_holder and len(self.copyright_holder) > config.MAX_COPYRIGHT_HOLDER_LENGTH:
            config.print_truncate("copyright_holder", self.license_id, self.copyright_holder)
            self.copyright_holder = self.copyright_holder[:config.MAX_COPYRIGHT_HOLDER_LENGTH]

    def as_dict(self):
        return {'license_id': self.license_id,
                'copyright_holder': self.copyright_holder,
                'description': self.description }

class CC_BYLicense(License):
    """
        The Attribution License lets others distribute, remix, tweak,
        and build upon your work, even commercially, as long as they credit
        you for the original creation. This is the most accommodating of
        licenses offered. Recommended for maximum dissemination and use of
        licensed materials.

        Reference: https://creativecommons.org/licenses/by/4.0
    """
    license_id = licenses.CC_BY

class CC_BY_SALicense(License):
    """
        The Attribution-ShareAlike License lets others remix, tweak, and
        build upon your work even for commercial purposes, as long as they
        credit you and license their new creations under the identical terms.
        This license is often compared to "copyleft" free and open source
        software licenses. All new works based on yours will carry the same
        license, so any derivatives will also allow commercial use. This is
        the license used by Wikipedia, and is recommended for materials that
        would benefit from incorporating content from Wikipedia and similarly
        licensed projects.

        Reference: https://creativecommons.org/licenses/by-sa/4.0
    """
    license_id = licenses.CC_BY_SA

class CC_BY_NDLicense(License):
    """
        The Attribution-NoDerivs License allows for redistribution, commercial
        and non-commercial, as long as it is passed along unchanged and in
        whole, with credit to you.

        Reference: https://creativecommons.org/licenses/by-nd/4.0
    """
    license_id = licenses.CC_BY_ND

class CC_BY_NCLicense(License):
    """
        The Attribution-NonCommercial License lets others remix, tweak, and
        build upon your work non-commercially, and although their new works
        must also acknowledge you and be non-commercial, they don't have to
        license their derivative works on the same terms.

        Reference: https://creativecommons.org/licenses/by-nc/4.0
    """
    license_id = licenses.CC_BY_NC

class CC_BY_NC_SALicense(License):
    """
        The Attribution-NonCommercial-ShareAlike License lets others remix, tweak,
        and build upon your work non-commercially, as long as they credit you and
        license their new creations under the identical terms.

        Reference: https://creativecommons.org/licenses/by-nc-sa/4.0
    """
    license_id = licenses.CC_BY_NC_SA

class CC_BY_NC_NDLicense(License):
    """
        The Attribution-NonCommercial-NoDerivs License is the most restrictive of
        our six main licenses, only allowing others to download your works and share
        them with others as long as they credit you, but they can't change them in
        any way or use them commercially.

        Reference: https://creativecommons.org/licenses/by-nc-nd/4.0
    """
    license_id = licenses.CC_BY_NC_ND

class AllRightsLicense(License):
    """
        The All Rights Reserved License indicates that the copyright holder reserves,
        or holds for their own use, all the rights provided by copyright law under
        one specific copyright treaty.

        Reference: http://www.allrights-reserved.com
    """
    license_id = licenses.ALL_RIGHTS_RESERVED

class PublicDomainLicense(License):
    """
        Public Domain work has been identified as being free of known restrictions
        under copyright law, including all related and neighboring rights.

        Reference: https://creativecommons.org/publicdomain/mark/1.0
    """
    require_copyright_holder = False
    license_id = licenses.PUBLIC_DOMAIN

class SpecialPermissionsLicense(License):
    """
        Special Permissions is a custom license to use when the current licenses
        do not apply to the content. The owner of this license is responsible for
        creating a description of what this license entails.
    """
    license_id = licenses.SPECIAL_PERMISSIONS

    def __init__(self, copyright_holder=None, description=None):
        assert description, "Special Permissions licenses must have a description"
        super(SpecialPermissionsLicense, self).__init__(copyright_holder=copyright_holder, description=description)



import os
import requests
import signal
import time
import urllib

from bs4 import BeautifulSoup
from selenium import webdriver
from urllib.parse import urlparse, unquote

from .caching import FileCache, CacheControlAdapter
from ricecooker.config import PHANTOMJS_PATH



# create a default session with basic caching mechanisms (similar to what a browser would do)
sess = requests.Session()
cache = FileCache('.webcache')
basic_adapter = CacheControlAdapter(cache=cache)
sess.mount('http://', basic_adapter)
sess.mount('https://', basic_adapter)

if PHANTOMJS_PATH is None:
    PHANTOMJS_PATH = os.path.join(os.getcwd(), "node_modules", "phantomjs-prebuilt", "bin", "phantomjs")

class WebDriver(object):

    def __init__(self, url, delay=1000):
        self.url = url
        self.delay = delay

    def __enter__(self):
        if not os.path.isfile(PHANTOMJS_PATH):
            raise Exception("You must install phantomjs-prebuilt in the directory"
                            " you're running in with `npm install phantomjs-prebuilt`"
                            " or set the environment variable `PHANTOMJS_PATH`")
        self.driver = webdriver.PhantomJS(executable_path=PHANTOMJS_PATH)
        self.driver.get(self.url)
        time.sleep(self.delay / 1000.0)
        return self.driver

    def __exit__(self ,type, value, traceback):
        # driver.quit() by itself doesn't suffice to fully terminate spawned
        # PhantomJS processes:
        # see https://github.com/seleniumhq/selenium/issues/767
        self.driver.service.process.send_signal(signal.SIGTERM)
        self.driver.quit()


def get_generated_html_from_driver(driver, tagname="html"):
    driver.execute_script("return document.getElementsByTagName('{tagname}')[0].innerHTML".format(tagname=tagname))


def calculate_relative_url(url, filename=None, baseurl=None, subpath=None):
    """
    Calculate the relative path for a URL relative to a base URL, possibly also injecting in a subpath prefix.
    """

    # ensure the provided subpath is a list
    if isinstance(subpath, str):
        subpath = subpath.strip("/").split("/")
    elif subpath is None:
        subpath = []

    # if a base path was supplied, calculate the file's subpath relative to it
    if baseurl:
        baseurl = urllib.parse.urljoin(baseurl, ".")  # ensure baseurl is normalized (to remove '/./' and '/../')
        assert url.startswith(baseurl), "URL must start with baseurl"
        subpath = subpath + url[len(baseurl):].strip("/").split("/")[:-1]

    # if we don't have a filename, extract it from the URL
    if not filename:
        filename = unquote(urlparse(url).path.split("/")[-1])

    # calculate the url path to use to refer to this file, relative to destpath
    relative_file_url = "/".join(["."] + subpath + [filename])

    return relative_file_url, subpath, filename


def download_file(url, destpath, filename=None, baseurl=None, subpath=None, middleware_callbacks=None, middleware_kwargs=None, request_fn=sess.get):
    """
    Download a file from a URL, into a destination folder, with optional use of relative paths and middleware processors.

    - If `filename` is set, that will be used as the name of the file when it's written to the destpath.
    - If `baseurl` is specified, the file will be put into subdirectory of destpath per the url's path relative to the baseurl.
    - If `subpath` is specified, it will be appended to destpath before deciding where to write the file.
    - If `middleware_callbacks` is specified, the returned content will be passed through those function(s) before being returned.
        - If `middleware_kwargs` are also specified, they will also be passed in to each function in middleware_callbacks.
    """

    relative_file_url, subpath, filename = calculate_relative_url(url, filename=filename, baseurl=baseurl, subpath=subpath)

    # ensure that the destination directory exists
    fulldestpath = os.path.join(destpath, *subpath)
    os.makedirs(fulldestpath, exist_ok=True)

    # make the actual request to the URL
    response = request_fn(url)
    content = response.content

    # if there are any middleware callbacks, apply them to the content
    if middleware_callbacks:
        content = content.decode()
        if not isinstance(middleware_callbacks, list):
            middleware_callbacks = [middleware_callbacks]
        kwargs = {
            "url": url,
            "destpath": destpath,
            "filename": filename,
            "baseurl": baseurl,
            "subpath": subpath,
            "fulldestpath": fulldestpath,
            "response": response,
        }
        kwargs.update(middleware_kwargs or {})
        for callback in middleware_callbacks:
            content = callback(content, **kwargs)

    # ensure content is encoded, as we're doing a binary write
    if isinstance(content, str):
        content = content.encode()

    # calculate the final destination for the file, and write the content out to there
    dest = os.path.join(fulldestpath, filename)
    with open(dest, "wb") as f:
        f.write(content)

    return relative_file_url, response



import argparse
import os
import pytest
import sys
import unittest

from docopt import docopt
from mock import patch

from ricecooker.exceptions import InvalidUsageException
from ricecooker.managers.progress import Status
from ricecooker.chefs import BaseChef

FAKE_CHEF_SCRIPT = 'chefs/fake_chef.py'


OLD_DOCOP_SPEC = """Usage: ricecooker uploadchannel [-huv] <file_path> [--warn] [--stage] [--compress] [--token=<t>] [--thumbnails] [--download-attempts=<n>] [--resume [--step=<step>] | --reset] [--prompt] [--publish] [[OPTIONS] ...]

Arguments:
  file_path        Path to file with channel data

Options:
  -h                          Help documentation
  -v                          Verbose mode
  -u                          Re-download files from file paths
  --warn                      Print out warnings to stderr
  --stage                     Stage updates rather than deploying them for manual verification on Kolibri Studio
  --compress                  Compress high resolution videos to low resolution videos
  --thumbnails                Automatically generate thumbnails for topics
  --token=<t>                 Authorization token (can be token or path to file with token) [default: #]
  --download-attempts=<n>     Maximum number of times to retry downloading files [default: 3]
  --resume                    Resume from ricecooker step (cannot be used with --reset flag)
  --step=<step>               Step to resume progress from (must be used with --resume flag) [default: LAST]
  --reset                     Restart session, overwriting previous session (cannot be used with --resume flag)
  --prompt                    Receive prompt to open the channel once it's uploaded
  --publish                   Automatically publish channel once it's been created
  [OPTIONS]                   Extra arguments to add to command line (e.g. key='field')

Steps (for restoring session):
  LAST (default):       Resume where the session left off
  INIT:                 Resume at beginning of session
  CONSTRUCT_CHANNEL:    Resume with call to construct channel
  CREATE_TREE:          Resume at set tree relationships
  DOWNLOAD_FILES:       Resume at beginning of download process
  GET_FILE_DIFF:        Resume at call to get file diff from Kolibri Studio
  START_UPLOAD:         Resume at beginning of uploading files to Kolibri Studio
  UPLOADING_FILES:      Resume at last upload request
  UPLOAD_CHANNEL:       Resume at beginning of uploading tree to Kolibri Studio
  PUBLISH_CHANNEL:      Resume at option to publish channel
  DONE:                 Resume at prompt to open channel

"""

@pytest.fixture
def fake_chef_path():
    testsdir = os.path.dirname(__file__)
    return os.path.join(testsdir, FAKE_CHEF_SCRIPT)

@pytest.fixture
def command_line_inputs(fake_chef_path):
    test_input_templates = [
      'ricecooker uploadchannel {} --token=letoken --resume --step=START_UPLOAD',
      'ricecooker uploadchannel {} --token=letoken --reset somethin=else extrakey=extraval',
      'ricecooker uploadchannel -uv {} --warn --compress --download-attempts=4 --token=besttokenever --resume --step=PUBLISH_CHANNEL --prompt --publish',
      'ricecooker uploadchannel {} -v --compress --token=katoken lang=en',
      'ricecooker uploadchannel {} -v --compress --token=katoken lang=en-us',
      'ricecooker uploadchannel {} -v --compress --token=katoken lang=fr-ca',
      'ricecooker uploadchannel {} -u --compress --token=katoken lang=en-us',
    ]
    test_inputs = [cmdt.format(fake_chef_path) for cmdt in test_input_templates]
    return test_inputs

def old_arguments_parser(cli_input):
    """
    Takes a string `cli_input` and parses it using the `docopt` module.
    Returns tuple of docopt arguments and extra kwargs (now called options).
    """
    test_argv = cli_input.split(' ')
    with patch.object(sys, 'argv', test_argv):
        arguments = docopt(OLD_DOCOP_SPEC)
    assert arguments is not None, 'doc opt parsing failed'

    # Parse OPTIONS for keyword arguments
    kwargs = {}
    for arg in arguments['OPTIONS']:
      try:
        kwarg = arg.split('=')
        kwargs.update({kwarg[0].strip(): kwarg[1].strip()})
      except IndexError:
        raise InvalidUsageException("Invalid kwarg '{0}' found: Must format as [key]=[value] (no whitespace)".format(arg))

    # Check if step is valid (if provided)
    step = arguments['--step']
    all_steps = [s.name for s in Status]
    if step.upper() not in all_steps:
      raise InvalidUsageException("Invalid step '{0}': Valid steps are {1}".format(step, all_steps))
    arguments['--step'] = step

    # Make sure max-retries can be cast as an integer
    try:
      int(arguments['--download-attempts'])
    except ValueError:
      raise InvalidUsageException("Invalid argument: Download-attempts must be an integer.")

    return arguments, kwargs


def arguments_to_args_renames(arguments):
    """
    The doc-opt parsed arguments include the - and -- and certain names are different.
    In order to have an apples-to-apples comparison with the new argparse approach,
    we must rename some of the keys.
    """
    args = {}
    for k, v in arguments.items():
        if k == '--download-attempts':
            args['download_attempts'] = int(v)
        elif k == '-u':
            args['update'] = v
        elif k == '-v':
            args['verbose'] = True # Should default to true
        elif k == '<file_path>':
            args['chef_script'] = v
        elif k == '-h':
            pass
        else:
            args[k.lstrip('-')] = v
    return args


def new_arg_parser(cli_input):
    """
    Takes a string `cli_input` and parses it using the BaseChef arg_parser.
    Returns tuple of args and options.
    """
    test_argv = cli_input.split(' ')
    with patch.object(sys, 'argv', test_argv):
        chef = BaseChef(compatibility_mode=True)
        args, options = chef.parse_args_and_options()
    assert args is not None, 'argparse parsing failed'
    return args, options

def dict_compare(d1, d2):
    """
    Compare two dicts. Usage: added, removed, modified, same = dict_compare(d1, d2)
    """
    d1_keys = set(d1.keys())
    d2_keys = set(d2.keys())
    intersect_keys = d1_keys.intersection(d2_keys)
    added = d1_keys - d2_keys
    removed = d2_keys - d1_keys
    modified = {o : (d1[o], d2[o]) for o in intersect_keys if d1[o] != d2[o]}
    same = set(o for o in intersect_keys if d1[o] == d2[o])
    return added, removed, modified, same




""" *********** CLI ARGUMENTS TESTS *********** """

def test_same_as_docopt(command_line_inputs):
    for line in command_line_inputs:
        arguments, kwargs = old_arguments_parser(line)
        orig_args = arguments_to_args_renames(arguments)
        del orig_args['uploadchannel']
        del orig_args['OPTIONS']

        args, options = new_arg_parser(line)
        del args['command']
        del args['quiet']  # new logging option was not present in docopt parser
        del args['debug']  # new logging option was not present in docopt parser

        added, removed, modified, same = dict_compare(orig_args, args)
        # print('different', added, removed, modified)
        assert orig_args == args, 'docopt arguments differ from argparse args'
        assert kwargs == options, 'extra key=value options differ'



import os
import re
import requests
import time
from urllib.parse import urlparse, parse_qs, urljoin
import uuid

from bs4 import BeautifulSoup
from selenium import webdriver
import selenium.webdriver.support.ui as selenium_ui
from requests_file import FileAdapter
from ricecooker.config import PHANTOMJS_PATH
from ricecooker.utils.html import download_file
from ricecooker.utils.caching import CacheForeverHeuristic, FileCache, CacheControlAdapter, InvalidatingCacheControlAdapter

DOWNLOAD_SESSION = requests.Session()                          # Session for downloading content from urls
DOWNLOAD_SESSION.mount('https://', requests.adapters.HTTPAdapter(max_retries=3))
DOWNLOAD_SESSION.mount('file://', FileAdapter())
cache = FileCache('.webcache')
forever_adapter= CacheControlAdapter(heuristic=CacheForeverHeuristic(), cache=cache)

DOWNLOAD_SESSION.mount('http://', forever_adapter)
DOWNLOAD_SESSION.mount('https://', forever_adapter)

headers = {
    "User-Agent": "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:20.0) Gecko/20100101 Firefox/20.0",
    "Accept-Encoding": "gzip, deflate",
    "Connection": "keep-alive"
}


def read(path, loadjs=False, session=None, driver=None, timeout=60,
        clear_cookies=True, loadjs_wait_time=3, loadjs_wait_for_callback=None):
    """Reads from source and returns contents
    
    Args:
        path: (str) url or local path to download
        loadjs: (boolean) indicates whether to load js (optional)
        session: (requests.Session) session to use to download (optional)
        driver: (selenium.webdriver) webdriver to use to download (optional)
        timeout: (int) Maximum number of seconds to wait for the request to complete.
        clear_cookies: (boolean) whether to clear cookies.
        loadjs_wait_time: (int) if loading JS, seconds to wait after the
            page has loaded before grabbing the page source
        loadjs_wait_for_callback: (function<selenium.webdriver>) if loading
            JS, a callback that will be invoked to determine when we can
            grab the page source. The callback will be called with the
            webdriver, and should return True when we're ready to grab the
            page source. For example, pass in an argument like:
            ``lambda driver: driver.find_element_by_id('list-container')``
            to wait for the #list-container element to be present before rendering.
    Returns: str content from file or page
    """
    session = session or DOWNLOAD_SESSION

    if clear_cookies:
        session.cookies.clear()

    try:
        if loadjs:                                              # Wait until js loads then return contents
            if PHANTOMJS_PATH:
                driver = driver or webdriver.PhantomJS(executable_path=PHANTOMJS_PATH)
            else:
                driver = driver or webdriver.PhantomJS()
            driver.get(path)
            if loadjs_wait_for_callback:
                selenium_ui.WebDriverWait(driver, 60).until(loadjs_wait_for_callback)
            time.sleep(loadjs_wait_time)
            return driver.page_source

        else:                                                   # Read page contents from url
            retry_count = 0
            max_retries = 5
            while True:
                try:
                    response = DOWNLOAD_SESSION.get(path, stream=True, timeout=timeout)
                    break
                except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
                    retry_count += 1
                    print("Error with connection ('{msg}'); about to perform retry {count} of {trymax}."
                        .format(msg=str(e), count=retry_count, trymax=max_retries))
                    time.sleep(retry_count * 1)
                    if retry_count >= max_retries:
                        raise e

            response.raise_for_status()
            return response.content

    except (requests.exceptions.MissingSchema, requests.exceptions.InvalidSchema):
        with open(path, 'rb') as fobj:                          # If path is a local file path, try to open the file
            return fobj.read()


def make_request(url, clear_cookies=True, timeout=60, *args, **kwargs):
    sess = DOWNLOAD_SESSION

    if clear_cookies:
        sess.cookies.clear()

    retry_count = 0
    max_retries = 5
    while True:
        try:
            response = sess.get(url, headers=headers, timeout=timeout, *args, **kwargs)
            break
        except (requests.exceptions.ConnectionError, requests.exceptions.ReadTimeout) as e:
            retry_count += 1
            print("Error with connection ('{msg}'); about to perform retry {count} of {trymax}."
                  .format(msg=str(e), count=retry_count, trymax=max_retries))
            time.sleep(retry_count * 1)
            if retry_count >= max_retries:
                raise e

    if response.status_code != 200:
        print("NOT FOUND:", url)

    return response


_CSS_URL_RE = re.compile(r"url\(['\"]?(.*?)['\"]?\)")


# TODO(davidhu): Use MD5 hash of URL (ideally file) instead.
def _derive_filename(url):
    name = os.path.basename(urlparse(url).path).replace('%', '_')
    return ("%s.%s" % (uuid.uuid4().hex, name)).lower()


def download_static_assets(doc, destination, base_url,
        request_fn=make_request, url_blacklist=[], js_middleware=None,
        css_middleware=None, derive_filename=_derive_filename):
    """
    Download all static assets referenced from an HTML page.
    The goal is to easily create HTML5 apps! Downloads JS, CSS, images, and
    audio clips.

    Args:
        doc: The HTML page source as a string or BeautifulSoup instance.
        destination: The folder to download the static assets to!
        base_url: The base URL where assets will be downloaded from.
        request_fn: The function to be called to make requests, passed to
            ricecooker.utils.html.download_file(). Pass in a custom one for custom
            caching logic.
        url_blacklist: A list of keywords of files to not include in downloading.
            Will do substring matching, so e.g. 'acorn.js' will match
            '/some/path/to/acorn.js'.
        js_middleware: If specificed, JS content will be passed into this callback
            which is expected to return JS content with any modifications.
        css_middleware: If specificed, CSS content will be passed into this callback
            which is expected to return CSS content with any modifications.

    Return the modified page HTML with links rewritten to the locations of the
    downloaded static files, as a BeautifulSoup object. (Call str() on it to
    extract the raw HTML.)
    """
    if not isinstance(doc, BeautifulSoup):
        doc = BeautifulSoup(doc, "html.parser")

    # Helper function to download all assets for a given CSS selector.
    def download_assets(selector, attr, url_middleware=None,
            content_middleware=None, node_filter=None):
        nodes = doc.select(selector)

        for i, node in enumerate(nodes):

            if node_filter:
                if not node_filter(node):
                    src = node[attr]
                    node[attr] = ''
                    print('        Skipping node with src ', src)
                    continue

            if node[attr].startswith('data:'):
                continue

            url = urljoin(base_url, node[attr])

            if _is_blacklisted(url, url_blacklist):
                print('        Skipping downloading blacklisted url', url)
                node[attr] = ""
                continue

            if url_middleware:
                url = url_middleware(url)

            filename = derive_filename(url)
            node[attr] = filename

            print("        Downloading", url, "to filename", filename)
            download_file(url, destination, request_fn=request_fn,
                    filename=filename, middleware_callbacks=content_middleware)

    def js_content_middleware(content, url, **kwargs):
        if js_middleware:
            content = js_middleware(content, url, **kwargs)

        # Polyfill localStorage and document.cookie as iframes can't access
        # them
        return (content
            .replace("localStorage", "_localStorage")
            .replace('document.cookie.split', '"".split')
            .replace('document.cookie', 'window._document_cookie'))

    def css_node_filter(node):
        return "stylesheet" in node["rel"]

    def css_content_middleware(content, url, **kwargs):
        if css_middleware:
            content = css_middleware(content, url, **kwargs)

        file_dir = os.path.dirname(urlparse(url).path)

        # Download linked fonts and images
        def repl(match):
            src = match.group(1)
            if src.startswith('//localhost'):
                return 'url()'
            # Don't download data: files
            if src.startswith('data:'):
                return match.group(0)
            src_url = urljoin(base_url, os.path.join(file_dir, src))

            if _is_blacklisted(src_url, url_blacklist):
                print('        Skipping downloading blacklisted url', src_url)
                return 'url()'

            derived_filename = derive_filename(src_url)
            download_file(src_url, destination, request_fn=request_fn,
                    filename=derived_filename)
            return 'url("%s")' % derived_filename

        return _CSS_URL_RE.sub(repl, content)

    # Download all linked static assets.
    download_assets("img[src]", "src")  # Images
    download_assets("link[href]", "href",
            content_middleware=css_content_middleware,
            node_filter=css_node_filter)  # CSS
    download_assets("script[src]", "src",
            content_middleware=js_content_middleware) # JS
    download_assets("source[src]", "src") # Potentially audio
    download_assets("source[srcset]", "srcset") # Potentially audio

    # ... and also run the middleware on CSS/JS embedded in the page source to
    # get linked files.
    for node in doc.select('style'):
        node.string = css_content_middleware(node.get_text(), url='')

    for node in doc.select('script'):
        if not node.attrs.get('src'):
            node.string = js_content_middleware(node.get_text(), url='')

    return doc


def _is_blacklisted(url, url_blacklist):
    return any((item in url.lower()) for item in url_blacklist)



""" Tests for exercise nodes, questions, and files """

import pytest
import uuid
import tempfile
from le_utils.constants import licenses, content_kinds, exercises
from ricecooker.classes.nodes import *


""" *********** EXERCISE FIXTURES *********** """
@pytest.fixture
def exercise_id():
    return "exercise-id"

@pytest.fixture
def channel_internal_domain():
    return "learningequality.org".encode('utf-8')

@pytest.fixture
def topic_node_id():
    return 'some-node-id'

@pytest.fixture
def exercise_content_id(channel_internal_domain, exercise_id):
    return uuid.uuid5(channel_internal_domain, exercise_id)

@pytest.fixture
def exercise_node_id(topic_node_id, exercise_content_id):
    return uuid.uuid5(topic_node_id, exercise_content_id.hex)

@pytest.fixture
def exercise_data(exercise_id):
    return {
    	"title": "exercise node test",
    	"description": None,
        "id" : exercise_id,
        "author": None,
        "license": licenses.PUBLIC_DOMAIN,
    }

@pytest.fixture
def exercise(exercise_data, channel_internal_domain, topic_node_id):
    node = ExerciseNode(
		source_id=exercise_data['id'],
		# description=exercise_data['description'],
		title=exercise_data['title'],
		author=exercise_data['author'],
		license=exercise_data['license'],
	)
    # node.set_ids(channel_internal_domain, topic_node_id)
    return node

@pytest.fixture
def exercise_json(exercise_data, exercise_content_id, exercise_node_id):
    return {
        "id" : exercise_data['id'],
        "title": exercise_data['title'],
        "description": "",
        "node_id": exercise_node_id.hex,
        "content_id": exercise_content_id.hex,
        "author": "",
        "children": [],
        "files": [],
        "kind": exercises.PERSEUS_QUESTION,
        "license": exercise_data['license'],
    }


""" *********** EXERCISE TESTS *********** """
def test_exercise_created(exercise):
    assert exercise is not None

def test_exercise_validate(exercise, exercise_data):
    assert exercise.source_id == exercise_data['id']
    assert exercise.title == exercise_data['title']
    # assert exercise.description == exercise_data['description']
    # assert exercise.author == exercise_data['author']
    # assert exercise.license == exercise_data['license']
    # assert exercise.kind == exercises.PERSEUS_QUESTION
#
# def test_exercise_to_dict(exercise):
#     assert exercise.default_preset == exercises.PERSEUS_QUESTION
#
# def test_exercise_add_question(exercise):
#     assert exercise.default_preset == exercises.PERSEUS_QUESTION
#
# def test_exercise_process_file(exercise):
#     assert exercise.default_preset == exercises.PERSEUS_QUESTION
#
# def test_exercise_process_exercise_data(exercise):
#     assert exercise.default_preset == exercises.PERSEUS_QUESTION


""" *********** BASE64FILE TESTS *********** """
def test_base64_process_file():
    assert True

def test_base64_validate():
    assert True

def test_base64_to_dict():
    assert True

def test_base64_convert_base64_to_file():
    assert True


""" *********** EXERCISEBASE64FILE TESTS *********** """
def test_exercisebase64_process_file():
    assert True

def test_exercisebase64_validate():
    assert True

def test_exercisebase64_to_dict():
    assert True

def test_exercisebase64_get_replacement_str():
    assert True


""" *********** EXERCISEIMAGEFILE TESTS *********** """
def test_exerciseimage_process_file():
    assert True

def test_exerciseimage_validate():
    assert True

def test_exerciseimage_to_dict():
    assert True

def test_exerciseimage_get_replacement_str():
    assert True


""" *********** EXERCISEGRAPHIEFILE TESTS *********** """
def test_exercisegraphie_process_file():
    assert True

def test_exercisegraphie_validate():
    assert True

def test_exercisegraphie_to_dict():
    assert True

def test_exercisegraphie_get_replacement_str():
    assert True

def test_exercisegraphie_generate_graphie_file():
    assert True


""" *********** QUESTION TESTS *********** """
def test_question_to_dict():
    assert True

def test_question_create_answer():
    assert True

def test_question_process_question():
    assert True

def test_question_set_images():
    assert True

def test_question_parse_html():
    assert True

def test_question_set_image():
    assert True

def test_question_validate():
    assert True


""" *********** PERSEUSQUESTION TESTS *********** """
def test_perseusquestion_to_dict():
    assert True

def test_perseusquestion_validate():
    assert True

def test_perseusquestion_process_question():
    assert True

def test_perseusquestion_process_image_field():
    assert True


""" *********** MULTIPLESELECTQUESTION TESTS *********** """
def test_multipleselectquestion_to_dict():
    assert True

def test_multipleselectquestion_validate():
    assert True


""" *********** SINGLESELECTQUESTION TESTS *********** """
def test_singleselectquestion_to_dict():
    assert True

def test_singleselectquestion_validate():
    assert True


""" *********** INPUTQUESTION TESTS *********** """
def test_inputquestion_to_dict():
    assert True

def test_inputquestion_validate():
    assert True




import json
import os

from ricecooker.classes import files, nodes, questions
from ricecooker.classes.licenses import get_license
from ricecooker.config import LOGGER
from ricecooker.exceptions import UnknownFileTypeError, UnknownQuestionTypeError
from ricecooker.classes.nodes import ChannelNode



# CONSTANTS USED TO SELECT APPROPRIATE CLASS DURING DESERIALIZATION FROM JSON
################################################################################

from le_utils.constants import content_kinds
TOPIC_NODE = content_kinds.TOPIC
VIDEO_NODE = content_kinds.VIDEO
AUDIO_NODE = content_kinds.AUDIO
EXERCISE_NODE = content_kinds.EXERCISE
DOCUMENT_NODE = content_kinds.DOCUMENT
HTML5_NODE = content_kinds.HTML5

# TODO(Ivan): add constants.file_types to le_utils and discuss with Jordan
# from le_utils.constants import file_types
VIDEO_FILE = "video"        # = file_types.VIDEO
AUDIO_FILE = "audio"        # = file_types.AUDIO
DOCUMENT_FILE = "document"  # = file_types.DOCUMENT etc..
HTML5_FILE = "html5"
THUMBNAIL_FILE = "thumbnail"
SUBTITLES_FILE = "subtitles"

from le_utils.constants import exercises
INPUT_QUESTION = exercises.INPUT_QUESTION
MULTIPLE_SELECTION = exercises.MULTIPLE_SELECTION
SINGLE_SELECTION = exercises.SINGLE_SELECTION
FREE_RESPONSE = exercises.FREE_RESPONSE
PERSEUS_QUESTION = exercises.PERSEUS_QUESTION




# JSON READ/WRITE HELPERS
################################################################################

def read_tree_from_json(srcpath):
    """
    Load ricecooker json tree data from json file at `srcpath`.
    """
    with open(srcpath) as infile:
        json_tree = json.load(infile)
        if json_tree is None:
            raise ValueError('Could not find ricecooker json tree')
    return json_tree

def write_tree_to_json_tree(destpath, json_tree):
    """
    Save contents of `json_tree` (dict) to json file at `destpath`.
    """
    parent_dir, _ = os.path.split(destpath)
    if not os.path.exists(parent_dir):
        os.makedirs(parent_dir, exist_ok=True)
    with open(destpath, 'w') as json_file:
        json.dump(json_tree, json_file, indent=2)



# CONSTRUCT CHANNEL FROM RICECOOKER JSON TREE
################################################################################

def get_channel_node_from_json(json_tree):
    """
    Build `ChannelNode` from json data provided in `json_tree`.
    """
    channel = ChannelNode(
        title=json_tree['title'],
        description=json_tree['description'],
        source_domain=json_tree['source_domain'],
        source_id=json_tree['source_id'],
        language=json_tree['language'],
        thumbnail=json_tree.get('thumbnail', None),
    )
    return channel

def build_tree_from_json(parent_node, sourcetree):
    """
    Recusively parse nodes in the list `sourcetree` and add them as children
    to the `parent_node`. Usually called with `parent_node` being a `ChannelNode`.
    """
    EXPECTED_NODE_TYPES = [TOPIC_NODE, VIDEO_NODE, AUDIO_NODE, EXERCISE_NODE,
                           DOCUMENT_NODE, HTML5_NODE]

    for source_node in sourcetree:
        kind = source_node['kind']
        if kind not in EXPECTED_NODE_TYPES:
            LOGGER.critical('Unexpected node type found: ' + kind)
            raise NotImplementedError('Unexpected node type found in json data.')

        if kind == TOPIC_NODE:
            child_node = nodes.TopicNode(
                source_id=source_node.get("source_id", None),
                title=source_node["title"],
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                thumbnail=source_node.get("thumbnail"),
            )
            parent_node.add_child(child_node)
            source_tree_children = source_node.get("children", [])
            build_tree_from_json(child_node, source_tree_children)

        elif kind == VIDEO_NODE:
            child_node = nodes.VideoNode(
                source_id=source_node["source_id"],
                title=source_node["title"],
                license=get_license(**source_node['license']),
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                derive_thumbnail=source_node.get('derive_thumbnail', True),  # video-specific option
                thumbnail=source_node.get('thumbnail'),
            )
            add_files(child_node, source_node.get("files") or [])
            parent_node.add_child(child_node)

        elif kind == AUDIO_NODE:
            child_node = nodes.AudioNode(
                source_id=source_node["source_id"],
                title=source_node["title"],
                license=get_license(**source_node['license']),
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                thumbnail=source_node.get('thumbnail'),
            )
            add_files(child_node, source_node.get("files") or [])
            parent_node.add_child(child_node)

        elif kind == EXERCISE_NODE:
            child_node = nodes.ExerciseNode(
                source_id=source_node["source_id"],
                title=source_node["title"],
                license=get_license(**source_node['license']),
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                thumbnail=source_node.get("thumbnail"),
                exercise_data=source_node.get("exercise_data"),
                questions=[],
            )
            add_questions(child_node, source_node.get("questions") or [])
            parent_node.add_child(child_node)

        elif kind == DOCUMENT_NODE:
            child_node = nodes.DocumentNode(
                source_id=source_node["source_id"],
                title=source_node["title"],
                license=get_license(**source_node['license']),
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                thumbnail=source_node.get("thumbnail"),
            )
            add_files(child_node, source_node.get("files") or [])
            parent_node.add_child(child_node)

        elif kind == HTML5_NODE:
            child_node = nodes.HTML5AppNode(
                source_id=source_node["source_id"],
                title=source_node["title"],
                license=get_license(**source_node['license']),
                author=source_node.get("author"),
                description=source_node.get("description"),
                language=source_node.get('language', None),
                thumbnail=source_node.get("thumbnail"),
            )
            add_files(child_node, source_node.get("files") or [])
            parent_node.add_child(child_node)

        else:
            LOGGER.critical("Encountered an unknown kind: " + str(source_node))
            continue

    return parent_node


def add_files(node, file_list):
    EXPECTED_FILE_TYPES = [VIDEO_FILE, AUDIO_FILE, DOCUMENT_FILE, HTML5_FILE,
                           THUMBNAIL_FILE, SUBTITLES_FILE]

    for f in file_list:
        file_type = f.get('file_type')
        if file_type not in EXPECTED_FILE_TYPES:
            LOGGER.critical(file_type)
            raise NotImplementedError('Unexpected File type found in channel json.')

        path = f.get('path')  # path can be an URL or a local path (or None)

        # handle different types of files
        if file_type == VIDEO_FILE:
            # handle three types of video files
            if 'youtube_id' in f:
                video_file = files.YouTubeVideoFile(
                    youtube_id=f['youtube_id'],
                    download_settings=f.get('download_settings', None),
                    high_resolution=f.get('high_resolution', True),
                    maxheight=f.get('maxheight', None),
                    language=f.get('language', None),
                )
            elif 'web_url' in f:
                video_file = files.WebVideoFile(
                    web_url=f['web_url'],
                    download_settings=f.get('download_settings', None),
                    high_resolution=f.get('high_resolution', True),
                    maxheight=f.get('maxheight', None),
                    language=f.get('language', None),
                )
            else:
                video_file = files.VideoFile(
                    path=f['path'],
                    language=f.get('language', None),
                    ffmpeg_settings=f.get('ffmpeg_settings'),
                )
            node.add_file(video_file)


        elif file_type == AUDIO_FILE:
            node.add_file(
                files.AudioFile(
                    path=f['path'],
                    language=f.get('language', None)
                )
            )

        elif file_type == DOCUMENT_FILE:
            node.add_file(
                files.DocumentFile(
                    path=path,
                    language=f.get('language', None)
                )
            )


        elif file_type == HTML5_FILE:
            node.add_file(
                files.HTMLZipFile(
                    path=path,
                    language=f.get('language', None)
                )
            )

        elif file_type == THUMBNAIL_FILE:
            if 'encoding' in f:
                node.add_file(
                    files.Base64ImageFile(
                        encoding=f['encoding'],
                    )
                )
            else:
                node.add_file(
                    files.ThumbnailFile(
                        path=path,
                        language=f.get('language', None),
                    )
                )

        elif file_type == SUBTITLES_FILE:
            if 'youtube_id' in f:
                node.add_file(
                    files.YouTubeSubtitleFile(
                        youtube_id=f['youtube_id'],
                        language=f['language']
                    )
                )
            else:
                node.add_file(
                    files.SubtitleFile(
                        path=path,
                        language=f['language']
                    )
                )

        else:
            raise UnknownFileTypeError("Unrecognized file type '{0}'".format(f['path']))




def add_questions(exercise_node, question_list):
    EXPECTED_QUESTION_TYPES = [INPUT_QUESTION, MULTIPLE_SELECTION, SINGLE_SELECTION,
                               FREE_RESPONSE, PERSEUS_QUESTION]

    for q in question_list:
        question_type = q.get('question_type')
        if question_type not in EXPECTED_QUESTION_TYPES:
            LOGGER.critical(question_type)
            raise NotImplementedError('Unexpected question type found in channel json.')

        question_text = q.get('question')
        hints = q.get('hints')
        hints = hints if isinstance(hints, str) else [hint for hint in hints or []]

        if question_type == exercises.MULTIPLE_SELECTION:
            q_obj = questions.MultipleSelectQuestion(
                id=q['id'],
                question=question_text,
                correct_answers=[answer for answer in q['correct_answers']],
                all_answers=[answer for answer in q['all_answers']],
                hints=hints,
            )
            exercise_node.add_question(q_obj)

        elif question_type == exercises.SINGLE_SELECTION:
            q_obj = questions.SingleSelectQuestion(
                id=q['id'],
                question=question_text,
                correct_answer=q['correct_answer'],
                all_answers=[answer for answer in q['all_answers']],
                hints=hints,
            )
            exercise_node.add_question(q_obj)

        elif question_type == exercises.INPUT_QUESTION:
            q_obj = questions.InputQuestion(
                id=q['id'],
                question=question_text,
                answers=[answer for answer in q['answers']],
                hints=hints,
            )
            exercise_node.add_question(q_obj)

        elif question_type == exercises.PERSEUS_QUESTION:
            q_obj = questions.PerseusQuestion(
                id=q['id'],
                raw_data=q.get('item_data'),
                source_url="https://www.khanacademy.org/",
            )
            exercise_node.add_question(q_obj)

        else:
            raise UnknownQuestionTypeError("Unrecognized question type '{0}': accepted types are {1}".format(question_type, [key for key, value in exercises.question_choices]))



import json
import sys

from .. import config


class ChannelManager:
    """ Manager for handling channel tree structure and communicating to server

        Attributes:
            channel (Channel): channel that manager is handling
    """
    def __init__(self, channel):
        self.channel = channel  # Channel to process
        self.uploaded_files = []
        self.failed_node_builds = {}
        self.failed_uploads = {}

    def validate(self):
        """ validate: checks if tree structure is valid
            Args: None
            Returns: boolean indicating if tree is valid
        """
        return self.channel.test_tree()

    def process_tree(self, channel_node):
        """
        Returns a list of all file names associated with a tree. Profiling suggests using a global list with `extend`
        is faster than using a global set or deque.
        :param channel_node: Root node of the channel being processed
        :return: The list of unique file names in `channel_node`.
        """
        file_names = []
        self.process_tree_recur(file_names, channel_node)
        return [x for x in set(file_names) if x]  # Remove any duplicate or null files

    def process_tree_recur(self, file_names, node):
        """
        Adds the names of all the files associated with the sub-tree rooted by `node` to `file_names` in post-order.
        :param file_names: A global list containing all file names associated with a tree
        :param node: The root of the current sub-tree being processed
        :return: None.
        """
        # Process node's children
        for child_node in node.children:
            self.process_tree_recur(file_names, child_node)  # Call children first in case a tiled thumbnail is needed

        file_names.extend(node.process_files())


    def check_for_files_failed(self):
        """ check_for_files_failed: print any files that failed during download process
            Args: None
            Returns: None
        """
        if len(config.FAILED_FILES) > 0:
            config.LOGGER.error("   {} file(s) have failed to download".format(len(config.FAILED_FILES)))
            for f in config.FAILED_FILES:
                title = "{0} {id}".format(f.node.kind.capitalize(), id=f.node.source_id)\
                        if f.node else "{0} {id}".format("Question", id=f.assessment_item.source_id)
                file_identifier = f.__dict__
                if hasattr(f, 'path') and f.path:
                    file_identifier = f.path
                elif hasattr(f, 'youtube_url') and f.youtube_url:
                    file_identifier = f.youtube_url
                config.LOGGER.warning("\t{0}: {id} \n\t   {err}".format(title, id=file_identifier, err=f.error))
        else:
            config.LOGGER.info("   All files were successfully downloaded")

    def get_file_diff(self, files_to_diff):
        """ get_file_diff: retrieves list of files that do not exist on content curation server
            Args: None
            Returns: list of files that are not on server
        """
        file_diff_result = []
        chunks = [files_to_diff[x:x+1000] for x in range(0, len(files_to_diff), 1000)]
        file_count = 0
        total_count = len(files_to_diff)
        for chunk in chunks:
            response = config.SESSION.post(config.file_diff_url(), data=json.dumps(chunk))
            response.raise_for_status()
            file_diff_result += json.loads(response._content.decode("utf-8"))
            file_count += len(chunk)
            config.LOGGER.info("\tGot file diff for {0} out of {1} files".format(file_count, total_count))

        return file_diff_result

    def upload_files(self, file_list):
        """ upload_files: uploads files to server
            Args:
                file_list (str): list of files to upload
            Returns: None
        """
        counter = 0
        files_to_upload = list(set(file_list) - set(self.uploaded_files)) # In case restoring from previous session
        try:
            for f in files_to_upload:
                with open(config.get_storage_path(f), 'rb') as file_obj:
                    response = config.SESSION.post(config.file_upload_url(), files={'file': file_obj})
                    if response.status_code == 200:
                        response.raise_for_status()
                        self.uploaded_files.append(f)
                        counter += 1
                        config.LOGGER.info("\tUploaded {0} ({count}/{total}) ".format(f, count=counter, total=len(files_to_upload)))
                    else:
                        self.failed_uploads[f] = response._content.decode('utf-8')
        finally:
            config.PROGRESS_MANAGER.set_uploading(self.uploaded_files)

    def reattempt_upload_fails(self):
        """ reattempt_upload_fails: uploads failed files to server
            Args: None
            Returns: None
        """
        if len(self.failed_uploads) > 0:
            config.LOGGER.info("\nReattempting to upload {0} file(s)...".format(len(self.failed_uploads)))
            current_fails = [k for k in self.failed_uploads]
            self.failed_uploads = {}
            self.upload_files(current_fails)

    def upload_channel_structure(self):
        from datetime import datetime
        start_time = datetime.now()
        config.LOGGER.info('   Uploading structure of channel {0}'.format(self.channel.title))

        channel_structure = {}
        self.fill_channel_structure(channel_structure, self.channel, 0)
        payload = {
            'channel_id': self.channel.to_dict()['id'],
            'channel_structure': channel_structure,
            'stage': config.STAGE,
        }
        response = config.SESSION.post(config.channel_structure_upload_url(), data=json.dumps(payload, sort_keys=False))
        response.raise_for_status()

        new_channel = json.loads(response._content.decode('utf-8'))
        end_time = datetime.now()
        config.LOGGER.info("Upload time: {time}s".format(time=(end_time - start_time).total_seconds()))
        return new_channel['channel_id'], config.open_channel_url(new_channel['channel_id'])

    def fill_channel_structure(self, cur_dict, cur_node, sort_order):
        children_dict = {}
        child_sort_order = 0
        for child in cur_node.children:
            self.fill_channel_structure(children_dict, child)
        cur_dict[cur_node.hashed_file_name] = children_dict

    def upload_tree(self):
        """ upload_tree: sends processed channel data to server to create tree
            Args: None
            Returns: link to uploadedchannel
        """
        from datetime import datetime
        start_time = datetime.now()
        root, channel_id = self.add_channel()
        self.node_count_dict = {"upload_count": 0, "total_count": self.channel.count()}

        config.LOGGER.info("\tPreparing fields...")
        self.truncate_fields(self.channel)

        self.add_nodes(root, self.channel)
        if self.check_failed(print_warning=False):
            failed = self.failed_node_builds
            self.failed_node_builds = {}
            self.reattempt_failed(failed)
            self.check_failed()
        channel_id, channel_link = self.commit_channel(channel_id)
        end_time = datetime.now()
        config.LOGGER.info("Upload time: {time}s".format(time=(end_time - start_time).total_seconds()))
        return channel_id, channel_link

    def truncate_fields(self, node):
        node.truncate_fields()
        for child in node.children:
            self.truncate_fields(child)

    def reattempt_failed(self, failed):
        for node_id in failed:
            node = failed[node_id]
            config.LOGGER.info("\tReattempting {0}s".format(str(node['node'])))
            for f in node['node'].files:
                # Attempt to upload file
                try:
                    assert f.filename, "File failed to download (cannot be uploaded)"
                    with open(config.get_storage_path(f.filename), 'rb') as file_obj:
                        response = config.SESSION.post(config.file_upload_url(), files={'file': file_obj})
                        response.raise_for_status()
                        self.uploaded_files.append(f.filename)
                except AssertionError as ae:
                    config.LOGGER.warning(ae)
            # Attempt to create node
            self.add_nodes(node_id, node['node'])

    def check_failed(self, print_warning=True):
        if len(self.failed_node_builds) > 0:
            if print_warning:
                config.LOGGER.warning("WARNING: The following nodes have one or more descendants that could not be created:")
                for node_id in self.failed_node_builds:
                    node = self.failed_node_builds[node_id]
                    config.LOGGER.warning("\t{} ({})".format(str(node['node']), node['error']))
            else:
                config.LOGGER.error("Failed to create descendants for {} node(s).".format(len(self.failed_node_builds)))
            return True
        else:
            config.LOGGER.info("   All nodes were created successfully.")
        return False

    def add_channel(self):
        """ add_channel: sends processed channel data to server to create tree
            Args: None
            Returns: link to uploadedchannel
        """
        config.LOGGER.info("   Creating channel {0}".format(self.channel.title))
        payload = {
            "channel_data":self.channel.to_dict(),
        }
        response = config.SESSION.post(config.create_channel_url(), data=json.dumps(payload))
        response.raise_for_status()
        new_channel = json.loads(response._content.decode("utf-8"))

        return new_channel['root'], new_channel['channel_id']

    def add_nodes(self, root_id, current_node, indent=1):
        """ add_nodes: adds processed nodes to tree
            Args:
                root_id (str): id of parent node on Kolibri Studio
                current_node (Node): node to publish children
                indent (int): level of indentation for printing
            Returns: link to uploadedchannel
        """
        # if the current node has no children, no need to continue
        if not current_node.children:
            return

        config.LOGGER.info("({count} of {total} uploaded) {indent}Processing {title} ({kind})".format(
            count=self.node_count_dict['upload_count'],
            total=self.node_count_dict['total_count'],
            indent="   " * indent,
            title=current_node.title,
            kind=current_node.__class__.__name__)
        )

        # Send children in chunks to avoid gateway errors
        try:
            chunks = [current_node.children[x:x+10] for x in range(0, len(current_node.children), 10)]
            for chunk in chunks:
                payload_children = []

                for child in chunk:
                    failed = [f for f in child.files if f.is_primary and (not f.filename or self.failed_uploads.get(f.filename))]
                    if any(failed):
                        if not self.failed_node_builds.get(root_id):
                            error_message = ""
                            for fail in failed:
                                reason = fail.filename + ": " + self.failed_uploads.get(fail.filename) if fail.filename else "File failed to download"
                                error_message = error_message + reason + ", "
                            self.failed_node_builds[root_id] = {'node': current_node, 'error': error_message[:-2]}
                    else:
                        payload_children.append(child.to_dict())
                payload = {
                    'root_id': root_id,
                    'content_data': payload_children
                }

                # When iceqube is integrated, use this method to utilize upload file optimizations
                # response = config.SESSION.post(config.add_nodes_from_file_url(), files={'file': json.dumps(payload)})

                response = config.SESSION.post(config.add_nodes_url(), data=json.dumps(payload))
                if response.status_code != 200:
                    self.failed_node_builds[root_id] = {'node': current_node, 'error': response.reason}
                else:
                    response_json = json.loads(response._content.decode("utf-8"))
                    self.node_count_dict['upload_count'] += len(chunk)

                    if response_json['root_ids'].get(child.get_node_id().hex):
                        for child in chunk:
                            self.add_nodes(response_json['root_ids'].get(child.get_node_id().hex), child, indent + 1)
        except ConnectionError as ce:
            self.failed_node_builds[root_id] = {'node': current_node, 'error': ce}

    def commit_channel(self, channel_id):
        """ commit_channel: commits channel to Kolibri Studio
            Args:
                channel_id (str): channel's id on Kolibri Studio
            Returns: channel id and link to uploadedchannel
        """
        payload = {
            "channel_id":channel_id,
            "stage": config.STAGE,
        }
        response = config.SESSION.post(config.finish_channel_url(), data=json.dumps(payload))
        if response.status_code != 200:
            config.LOGGER.error("\n\nCould not activate channel: {}\n".format(response._content.decode('utf-8')))
            if response.status_code == 403:
                config.LOGGER.error("Channel can be viewed at {}\n\n".format(config.open_channel_url(channel_id, staging=True)))
                sys.exit()
        response.raise_for_status()
        new_channel = json.loads(response._content.decode("utf-8"))
        channel_link = config.open_channel_url(new_channel['new_channel'])
        return channel_id, channel_link

    def publish(self, channel_id):
        """ publish: publishes tree to Kolibri
            Args:
                channel_id (str): channel's id on Kolibri Studio
            Returns: None
        """
        payload = {
            "channel_id":channel_id,
        }
        response = config.SESSION.post(config.publish_channel_url(), data=json.dumps(payload))
        response.raise_for_status()




class PathBuilder:
    """
    Class for formatting paths to write to DataWriter.
    """

    path = None             # List of items in path
    channel_name = None     # Name of channel

    def __init__(self, channel_name=None):
        """ Args: channel_name: (str) Name of channel """
        self.channel_name = channel_name or "Channel"
        self.path = [self.channel_name]

    def __str__(self):
        """Converts path list to string
        e.g. [Channel, Topic, Subtopic] -> Channel/Topic/Subtopic
        Returns: str path
        """
        return "/".join(self.path)

    def reset(self):
        """Clear path
        Args: None
        Returns: None
        """
        self.path = [self.channel_name]

    def set(self, *path):
        """Set path from root
        Args: path: (str) items to add to path
        Returns: None
        """
        self.path = [self.channel_name]
        self.path.extend(list(path))

    def open_folder(self, path_item):
        """Add item to path
        Args: path_item: (str) item to add to path
        Returns: None
        """
        self.path.append(path_item)

    def go_to_parent_folder(self):
        """Go back one level in path
        Args: None
        Returns: last item in path
        """
        if len(self.path) > 1:
            return self.path.pop()



import requests
import cachecontrol

from datetime import datetime, timedelta
from email.utils import parsedate

from cachecontrol import CacheControlAdapter
from cachecontrol.caches.file_cache import FileCache
from cachecontrol.heuristics import BaseHeuristic, expire_after, datetime_to_header


class NeverCache(BaseHeuristic):
    """
    Don't cache the response at all.
    """
    def update_headers(self, response):
        return {'cache-control': 'no-cache'}


class CacheForeverHeuristic(BaseHeuristic):
    """
    Cache the response effectively forever.
    """
    def update_headers(self, response):
        headers = {}
        expires = expire_after(timedelta(weeks=10*52), date=datetime.now())
        headers['expires'] = datetime_to_header(expires)
        headers['cache-control'] = 'public'
        
        return headers


class InvalidatingCacheControlAdapter(CacheControlAdapter):
    """
    Cache control adapter that deletes items from the cache as they're requested.
    Default heuristic is also set to a non-caching heuristic.
    """

    def __init__(self, heuristic=None, *args, **kw):
        if not heuristic:
            heuristic = NeverCache()
        super(InvalidatingCacheControlAdapter, self).__init__(*args, heuristic=heuristic, **kw)

    def send(self, request, **kw):

        # delete any existing cached value from the cache
        try:
            cache_url = self.controller.cache_url(request.url)
            self.cache.delete(cache_url)
        except FileNotFoundError:
            pass

        resp = super(InvalidatingCacheControlAdapter, self).send(request, **kw)

        return resp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import ricecooker
from setuptools import setup, find_packages


try:
    import pypandoc
    readme = pypandoc.convert_file('README.md', 'rst')
except (IOError, ImportError):
    readme = open('README.md').read()

with open('docs/history.rst') as history_file:
    history = history_file.read()

requirements = [
    "pytest>=3.0.2",
    "requests>=2.11.1",
    "pillow>=3.3.1",
    "docopt>=0.6.2",
    "le_utils>=0.0.9rc24",
    "validators",
    "requests_file",
    "beautifulsoup4==4.5.1",
    "pressurecooker>=0.0.16",
    "selenium==3.0.1",
    "youtube-dl",
    "html5lib",
    "cachecontrol==0.12.0",
    "lockfile==0.12.2",
    "css-html-js-minify==2.2.2",
    "websocket-client==0.40.0",
    "mock==2.0.0",
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='ricecooker',
    version=ricecooker.__version__,
    description="API for adding content to the Kolibri content curation server",
    long_description=readme + '\n\n' + history,
    author="Learning Equality",
    author_email='dev@learningequality.org',
    url='https://github.com/learningequality/ricecooker',
    packages=find_packages(),
    package_dir={'ricecooker':'ricecooker'},
    entry_points={},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='ricecooker',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Education',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



# Settings for rice cooker
import hashlib
import logging
import os

import requests
from requests_file import FileAdapter

UPDATE = False
COMPRESS = False
THUMBNAILS = False
PUBLISH = False
PROGRESS_MANAGER = None
SUSHI_BAR_CLIENT = None
STAGE = False
LOGGER = logging.getLogger()

# Domain and file store location for uploading to production Studio server
DOMAIN_ENV = os.getenv('STUDIO_URL', None)
if DOMAIN_ENV is None:  # check old ENV varable for backward compatibility
    DOMAIN_ENV = os.getenv('CONTENTWORKSHOP_URL', None)
DOMAIN = DOMAIN_ENV if DOMAIN_ENV else "https://contentworkshop.learningequality.org"
if DOMAIN.endswith('/'):
    DOMAIN = DOMAIN.rstrip('/')
FILE_STORE_LOCATION =  hashlib.md5(DOMAIN.encode('utf-8')).hexdigest()

# Allow users to choose which phantomjs they use
PHANTOMJS_PATH = os.getenv('PHANTOMJS_PATH', None)

# URL for authenticating user on Kolibri Studio
AUTHENTICATION_URL = "{domain}/api/internal/authenticate_user_internal"

# URL for checking compatible version on Kolibri Studio
VERSION_CHECK_URL = "{domain}/api/internal/check_version"

# URL for getting file diff
FILE_DIFF_URL = "{domain}/api/internal/file_diff"

# URL for uploading files to server
FILE_UPLOAD_URL = "{domain}/api/internal/file_upload"

# URL for uploading channel structure to server
CHANNEL_STRUCTURE_UPLOAD_URL = "{domain}/api/internal/channel_structure_upload"

# URL for creating channel on server
CREATE_CHANNEL_URL = "{domain}/api/internal/create_channel"

# URL for adding nodes to channel
ADD_NODES_URL = "{domain}/api/internal/add_nodes"

# URL for adding nodes to channel from file
ADD_NODES_FROM_FILE_URL = "{domain}/api/internal/api_add_nodes_from_file"

# URL for making final changes to channel
FINISH_CHANNEL_URL = "{domain}/api/internal/finish_channel"

# URL to return after channel is created
OPEN_CHANNEL_URL = "{domain}/channels/{channel_id}/{access}"

# URL for publishing channel
PUBLISH_CHANNEL_URL = "{domain}/api/internal/publish_channel"

# Folder to store downloaded files
STORAGE_DIRECTORY = "storage"

# Folder to store progress tracking information
RESTORE_DIRECTORY = "restore"

# Session for communicating to Kolibri Studio
SESSION = requests.Session()

# Cache for filenames
FILECACHE_DIRECTORY = ".ricecookerfilecache"

FAILED_FILES = []

# Session for downloading files
DOWNLOAD_SESSION = requests.Session()
DOWNLOAD_SESSION.mount('file://', FileAdapter())

# Sushi bar server
SUSHIBAR_URL = os.getenv('SUSHIBAR_URL', "https://sushibar.learningequality.org")
if SUSHIBAR_URL.endswith('/'):
    SUSHIBAR_URL = SUSHIBAR_URL.rstrip('/')
if not SUSHIBAR_URL.startswith('http'):
   SUSHIBAR_URL = 'https://' + SUSHIBAR_URL        # in case only hostname given
SUSHI_BAR_HTTP = SUSHIBAR_URL
SUSHI_BAR_WEBSOCKET = SUSHIBAR_URL.replace('http', 'ws', 1)
SUSHI_BAR_CHANNEL_URL = "{domain}/api/channels/"
SUSHI_BAR_CHANNEL_RUNS_URL = "{domain}/api/channelruns/"
SUSHI_BAR_CHANNEL_RUNS_DETAIL_URL = "{domain}/api/channelruns/{run_id}/"
SUSHI_BAR_STAGES_URL = "{domain}/api/channelruns/{run_id}/stages/"
SUSHI_BAR_PROGRESS_URL = "{domain}/api/channelruns/{run_id}/progress/"
SUSHI_BAR_LOGS_URL = "{domain}/logs/{run_id}/"
SUSHI_BAR_CONTROL_URL = "{domain}/control/{channel_id}/"


# Character limits based on Kolibri models
TRUNCATE_MSG = "\t\t{kind} {id}: {field} {value} is too long - max {max} characters (truncating)"

MAX_TITLE_LENGTH = 200
MAX_SOURCE_ID_LENGTH = 200
MAX_DESCRIPTION_LENGTH = 400
MAX_AUTHOR_LENGTH = 200
MAX_SOURCE_URL_LENGTH = 400
MAX_ORIGINAL_FILENAME_LENGTH = 255
MAX_LICENSE_DESCRIPTION_LENGTH = 400
MAX_COPYRIGHT_HOLDER_LENGTH = 200

MAX_CHAR_LIMITS = {
    "title": {
        "kind": "Node",
        "field": "title",
        "max": MAX_TITLE_LENGTH
    },
    "source_id": {
        "kind": "Node",
        "field": "source_id",
        "max": MAX_SOURCE_ID_LENGTH
    },
    "description": {
        "kind": "Node",
        "field": "description",
        "max": MAX_DESCRIPTION_LENGTH
    },
    "author": {
        "kind": "Node",
        "field": "source_id",
        "max": MAX_AUTHOR_LENGTH
    },
    "question_source_url": {
        "kind": "Question",
        "field": "source url",
        "max": MAX_SOURCE_URL_LENGTH
    },
    "original_filename": {
        "kind": "File",
        "field": "original filename",
        "max": MAX_ORIGINAL_FILENAME_LENGTH
    },
    "file_source_url": {
        "kind": "File",
        "field": "source url",
        "max": MAX_SOURCE_URL_LENGTH
    },
    "license_description": {
        "kind": "License",
        "field": "license description",
        "max": MAX_LICENSE_DESCRIPTION_LENGTH
    },
    "copyright_holder": {
        "kind": "License",
        "field": "copyright holder",
        "max": MAX_COPYRIGHT_HOLDER_LENGTH
    },
}


def print_truncate(field, id, value, kind=None):
    limit = MAX_CHAR_LIMITS.get(field)
    LOGGER.warning(TRUNCATE_MSG.format(kind=kind or limit["kind"], id=id, field=limit["field"], value=value, max=limit["max"]))

def get_storage_path(filename):
    """ get_storage_path: returns path to storage directory for downloading content
        Args: filename (str): Name of file to store
        Returns: string path to file
    """
    directory = os.path.join(STORAGE_DIRECTORY, filename[0], filename[1])
    # Make storage directory for downloaded files if it doesn't already exist
    if not os.path.exists(directory) :
        os.makedirs(directory)

    return os.path.join(directory, filename)

def authentication_url():
    """ authentication_url: returns url to login to Kolibri Studio
        Args: None
        Returns: string url to authenticate_user_internal endpoint
    """
    return AUTHENTICATION_URL.format(domain=DOMAIN)

def init_file_mapping_store():
    """ init_file_mapping_store: creates log to keep track of downloaded files
        Args: None
        Returns: None
    """
    # Make storage directory for restore files if it doesn't already exist
    path = os.path.join(RESTORE_DIRECTORY, FILE_STORE_LOCATION)
    if not os.path.exists(path):
        os.makedirs(path)

def get_restore_path(filename):
    """ get_restore_path: returns path to directory for restoration points
        Args:
            filename (str): Name of file to store
        Returns: string path to file
    """
    path = os.path.join(RESTORE_DIRECTORY, FILE_STORE_LOCATION)
    if not os.path.exists(path):
        os.makedirs(path)
    return os.path.join(path, filename + '.pickle')


def check_version_url():
    """ check_version_url: returns url to check ricecooker version
        Args: None
        Returns: string url to check version endpoint
    """
    return VERSION_CHECK_URL.format(domain=DOMAIN)


def file_diff_url():
    """ file_diff_url: returns url to get file diff
        Args: None
        Returns: string url to file_diff endpoint
    """
    return FILE_DIFF_URL.format(domain=DOMAIN)

def file_upload_url():
    """ file_upload_url: returns url to upload files
        Args: None
        Returns: string url to file_upload endpoint
    """
    return FILE_UPLOAD_URL.format(domain=DOMAIN)


def channel_structure_upload_url():
    """
    Returns a string representation of the API endpoint for uploading a channel's structure.
    """
    return CHANNEL_STRUCTURE_UPLOAD_URL.format(domain=DOMAIN)


def create_channel_url():
    """ create_channel_url: returns url to create channel
        Args: None
        Returns: string url to create_channel endpoint
    """
    return CREATE_CHANNEL_URL.format(domain=DOMAIN)

def add_nodes_url():
    """ add_nodes_url: returns url to add nodes to channel
        Args: None
        Returns: string url to add_nodes endpoint
    """
    return ADD_NODES_URL.format(domain=DOMAIN)

def add_nodes_from_file_url():
    """ add_nodes_from_file_url: returns url to add nodes to channel using json file
        Args: None
        Returns: string url to add_nodes endpoint
    """
    return ADD_NODES_FROM_FILE_URL.format(domain=DOMAIN)

def finish_channel_url():
    """ finish_channel_url: returns url to finish uploading a channel
        Args: None
        Returns: string url to finish_channel endpoint
    """
    return FINISH_CHANNEL_URL.format(domain=DOMAIN)

def open_channel_url(channel, staging=False):
    """ open_channel_url: returns url to uploaded channel
        Args:
            channel (str): channel id of uploaded channel
        Returns: string url to open channel
    """
    return OPEN_CHANNEL_URL.format(domain=DOMAIN, channel_id=channel, access='staging' if staging or STAGE else 'edit')

def publish_channel_url():
    """ open_channel_url: returns url to publish channel
        Args: None
        Returns: string url to publish channel
    """
    return PUBLISH_CHANNEL_URL.format(domain=DOMAIN)

def sushi_bar_channels_url():
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_CHANNEL_URL.format(domain=SUSHI_BAR_HTTP)

def sushi_bar_channel_runs_url():
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_CHANNEL_RUNS_URL.format(domain=SUSHI_BAR_HTTP)

def sushi_bar_channel_runs_detail_url(run_id):
    """
    Returns the url to patch a channel run.
    """
    return SUSHI_BAR_CHANNEL_RUNS_DETAIL_URL.format(domain=SUSHI_BAR_HTTP,
                                                    run_id=run_id)

def sushi_bar_stages_url(run_id):
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_STAGES_URL.format(domain=SUSHI_BAR_HTTP, run_id=run_id)

def sushi_bar_progress_url(run_id):
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_PROGRESS_URL.format(domain=SUSHI_BAR_HTTP, run_id=run_id)

def sushi_bar_logs_url(run_id):
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_LOGS_URL.format(domain=SUSHI_BAR_WEBSOCKET, run_id=run_id)

def sushi_bar_control_url(channel_id):
    """
    Returns the url to report the progress of a sushi chef
    """
    return SUSHI_BAR_CONTROL_URL.format(domain=SUSHI_BAR_WEBSOCKET, channel_id=channel_id)



import csv
import os
import zipfile
from io import StringIO
from ricecooker.utils.downloader import read
from le_utils.constants import licenses

DEFAULT_WRITE_TO_PATH = "Channel.zip"
NO_COPYRIGHT_HOLDER_REQUIRED = [licenses.PUBLIC_DOMAIN]

class DataWriter():
    """
        Class for writing topic tree to standardized folder/csv structure
    """

    zf = None               # Zip file to write to
    write_to_path = None    # Where to write zip file

    def __init__(self, write_to_path=DEFAULT_WRITE_TO_PATH):
        """ Args: write_to_path: (str) where to write zip file (optional) """
        self.map = {}                       # Keeps track of content to write to csv
        self.write_to_path = write_to_path  # Where to write zip file

    def __enter__(self):
        """ Called when opening context (e.g. with DataWriter() as writer: ) """
        self.open()
        return self

    def __exit__(self, type, value, traceback):
        """ Called when closing context """
        self.close()

    def _write_to_zip(self, path, contents):
        """ _write_to_zip: Write file to zip
            Args:
                path: (str) where in zip to write file
                contents: (str) contents of file to write
            Returns: None
        """
        if isinstance(path, list):
            path = os.path.sep.join(path)
        self.zf.writestr(path, contents)

    def _commit(self, path, title, source_id=None, description=None, author=None, language=None, license=None, copyright_holder=None, license_description=None, thumbnail=None):
        """ _commit: Adds folder/file to map
            Args:
                path: (str) where in zip to write file
                title: (str) content's title
                source_id: (str) content's original id (optional)
                description: (str) description of content (optional)
                author (str): who created the content (optional)
                language (str): language of content (optional)
                license: (str) content's license (optional)
                copyright_holder (str): holder of content's license (optional)
                license_description (str): description of content's license (optional)
                thumbnail (str):  path to thumbnail in zip (optional)
            Returns: None
        """
        node = self.map.get(path)
        self.map.update({path: {
            'title': title or node and node.get('path'),
            'source_id': source_id or node and node.get('source_id'),
            'description': description or node and node.get('description'),
            'author': author or node and node.get('author'),
            'language': language or node and node.get('language'),
            'license': license or node and node.get('license'),
            'copyright_holder': copyright_holder or node and node.get('copyright_holder'),
            'license_description': license_description or node and node.get('license_description'),
            'thumbnail': thumbnail or node and node.get('thumbnail'),
        }})

    def _parse_path(self, path):
        """ _parse_path: Go through path and make sure topics exist in map
            Args: path: (str) path to file in zip
            Returns: None
        """
        paths = path.split('/')
        current_path = paths[0]
        for p in paths[1:]:
            current_path = "{}/{}".format(current_path, p)
            if not self.map.get(current_path): # Create any folders that might not exist yet
                self._commit(current_path, p)

    def _write_metadata(self):
        """ _write_metadata: Writes node metadata to csv file
            Args: None
            Returns: None
        """
        string_buffer = StringIO()
        writer = csv.writer(string_buffer, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(['Path *', 'Title *', 'Source ID', 'Description', 'Author', 'Language', 'License ID *', 'License Description', 'Copyright Holder', 'Thumbnail'])
        for k in self.map:
            node = self.map[k]
            writer.writerow([k, node['title'], node['source_id'], node['description'], node['author'], node['language'], node['license'], node['license_description'], node['copyright_holder'], node['thumbnail']])

        self.zf.writestr('Content.csv', string_buffer.getvalue())


    """ USER-FACING METHODS """

    def open(self):
        """ open: Opens zipfile to write to
            Args: None
            Returns: None
        """
        self.zf = zipfile.ZipFile(self.write_to_path, "w")

    def close(self):
        """ close: Close zipfile when done
            Args: None
            Returns: None
        """
        self._write_metadata()
        self.zf.close()

    def add_channel(self, title, source_id, domain, language, description=None, thumbnail=None):
        """ add_channel: Creates channel metadata
            Args:
                source_id: (str) channel's unique id
                domain: (str) who is providing the content (e.g. learningequality.org)
                title: (str): name of channel
                language: (str): language code for channel (e.g. 'en')
                description: (str) description of the channel (optional)
                thumbnail: (str) path to thumbnail in zip (optional)
            Returns: None
        """
        string_buffer = StringIO()
        writer = csv.writer(string_buffer, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(['Title', 'Description', 'Domain', 'Source ID', 'Language', 'Thumbnail'])
        writer.writerow([title, description, domain, source_id, language, thumbnail])
        self.zf.writestr('Channel.csv', string_buffer.getvalue())

    def add_folder(self, path, title, description=None, language=None, thumbnail=None, source_id=None, **node_data):
        """ add_folder: Creates folder in csv
            Args:
                path: (str) where in zip to write folder
                title: (str) content's title
                source_id: (str) content's original id (optional)
                description: (str) description of content (optional)
                language (str): language of content (optional)
                thumbnail (str):  path to thumbnail in zip (optional)
            Returns: None
        """
        self._parse_path(path)
        path = path if path.endswith(title) else "{}/{}".format(path, title)
        self._commit(path, title, description=description, language=language, thumbnail=thumbnail, source_id=source_id)

    def add_file(self, path, title, download_url, write_data=True, ext=None, license=None, copyright_holder=None, **node_data):
        """ add_file: Creates file in csv and writes file to zip
            Args:
                path: (str) where in zip to write file
                title: (str) content's title
                download_url: (str) url or local path to download from
                write_data: (boolean) indicates whether to add as a csv entry (optional)
                ext: (str) extension to use for file
                license (str): content's license
                copyright_holder (str): holder of content's license (required except for PUBLIC_DOMAIN)
                license_description (str): description of content's license (optional)
                source_id: (str) content's original id (optional)
                description: (str) description of content (optional)
                author (str): who created the content (optional)
                language (str): language of content (optional)
                thumbnail (str):  path to thumbnail in zip (optional)
            Returns: path to file in zip
        """
        if write_data:
            assert license, "Files must have a license"
            copyright_holder = None if not copyright_holder or copyright_holder.strip() == '' else copyright_holder
            assert license in NO_COPYRIGHT_HOLDER_REQUIRED or copyright_holder, "Licenses must have a copyright holder if they are not public domain"

        self._parse_path(path)
        if not ext:
            _name, ext = os.path.splitext(download_url or "")
            ext = ext.lower()  # normalize to lowercase extensions inside zip archive
        filepath = "{}/{}{}".format(path, title, ext)
        if download_url and filepath:
            self._write_to_zip(filepath, read(download_url))
            if write_data:
                self._commit(filepath, title, license=license, copyright_holder=copyright_holder, **node_data)
            return filepath



""" Tests for CSV exercises channel logic """
import os
import pytest
import tempfile

from ricecooker.chefs import LineCook
from ricecooker.utils.jsontrees import read_tree_from_json
from ricecooker.utils.metadata_provider import CsvMetadataProvider



@pytest.fixture
def channeldir():
    return 'tests/testchannels/csv_channel_with_exercises/channeldir'


def test_exercises_metadata_provider(channeldir):
    _, channeldirname = os.path.split(channeldir)
    mp = CsvMetadataProvider(channeldir)
    assert mp is not None, 'CsvMetadataProvider does not exist'
    mp.validate_headers()
    assert mp.has_exercises(), 'has exercises'
    assert mp.get_channel_info()['source_id'] == 'csv_channel_with_exercises', 'check source id'    
    #
    assert len(mp.contentcache.keys()) == 8, 'Found too many items'
    assert len(mp.get_exercises_for_dir((channeldirname,))) == 1, 'one exercise in root'
    assert len(mp.get_exercises_for_dir((channeldirname,'exercises'))) == 3, '3 exercise in exercises/'


def test_exercises_linecook(channeldir):
    tmpdir_path = tempfile.mkdtemp()

    linecook = LineCook()
    linecook.TREES_DATA_DIR = tmpdir_path
    linecook.RICECOOKER_JSON_TREE = 'test_ricecooker_json_tree.json'

    args = dict(
        channeldir=channeldir,
        channelinfo='Channel.csv',
        contentinfo='Content.csv',
        exercisesinfo='Exercises.csv',
        questionsinfo='ExerciseQuestions.csv',
        token='???',
    )
    options = {}
    linecook.pre_run(args, options)
    
    jsontree_path = os.path.join(tmpdir_path, linecook.RICECOOKER_JSON_TREE)
    assert os.path.exists(jsontree_path), 'output json exists'
    test_tree = read_tree_from_json(jsontree_path)
    assert len(test_tree['children']) == 3, 'exercise node + two dirs'

    # cleanup
    os.remove(jsontree_path)
    os.rmdir(tmpdir_path)


